# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.Qsci import QsciLexerPython
from qtpy.QtGui import QColor, QFont

from karabogui.fonts import get_qfont

# Collection of comment types as Enum.
COMMENTS = (
    QsciLexerPython.Comment, QsciLexerPython.CommentBlock,
    QsciLexerPython.TripleDoubleQuotedString,
    QsciLexerPython.TripleSingleQuotedString
)

# Syntax color map
COLOR_MAP = {
    QsciLexerPython.Default: QColor("#000000"),
    QsciLexerPython.Comment: QColor("#8C8C8C"),
    QsciLexerPython.Number: QColor("#1750EB"),
    QsciLexerPython.DoubleQuotedString: QColor("#067D17"),
    QsciLexerPython.DoubleQuotedFString: QColor("#067D17"),
    QsciLexerPython.SingleQuotedString: QColor("#067D17"),
    QsciLexerPython.SingleQuotedFString: QColor("#067D17"),
    QsciLexerPython.UnclosedString: QColor("#067D17"),
    QsciLexerPython.Keyword: QColor("#0033B3"),
    QsciLexerPython.TripleSingleQuotedString: QColor("#8C8C8C"),
    QsciLexerPython.TripleDoubleQuotedString: QColor("#8C8C8C"),
    QsciLexerPython.ClassName: QColor("#000000"),
    QsciLexerPython.FunctionMethodName: QColor("#00627A"),
    QsciLexerPython.Operator: QColor("#000000"),
    QsciLexerPython.Identifier: QColor("#000000"),
    QsciLexerPython.CommentBlock: QColor("#8C8C8C"),
    QsciLexerPython.HighlightedIdentifier: QColor("#94558D"),
    QsciLexerPython.Decorator: QColor("#9E880D"),
}


class PythonLexer(QsciLexerPython):
    """
    Syntax highlighter for Macro Editor
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        # Set font
        font = get_qfont("Source Code Pro,10,-1,5,50,0,0,0,0,0")
        for item, color in COLOR_MAP.items():
            self.setColor(color, item)
            self.setFont(font, item)

        # Comment text in Italics
        comment_font = QFont(font)
        comment_font.setItalic(True)
        for comment in COMMENTS:
            self.setFont(comment_font, comment)

        # code fold
        self.setFoldComments(True)
        self.setFoldQuotes(True)
        self.setFoldCompact(True)

    def keywords(self, set_number):
        """
        Extend method of QsciLexerPython to create a keyword set for 'True',
        'False', 'self' and 'cls' to color them in the editor.
        """
        ret_value = super().keywords(set_number)
        # Other Python reserved keywords are defined as set number 1 in
        # QsciLexerPython
        if set_number == 2:
            ret_value = "True False self cls"
        return ret_value



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import random
from platform import system

import pytest
from qtpy.QtCore import QPoint, QSize
from qtpy.QtGui import QFont, QFontDatabase
from qtpy.QtWidgets import QWidget

from karabo.common.scenemodel.api import (
    SCENE_DEFAULT_DPI, SCENE_FONT_SIZE, SCENE_MAC_DPI, DeviceSceneLinkModel,
    LabelModel, SceneLinkModel, SceneModel, StickerModel, WebLinkModel)
from karabo.common.scenemodel.tests.utils import single_model_round_trip
from karabo.native import Configurable, VectorString
from karabogui.binding.api import (
    DeviceProxy, PropertyProxy, ProxyStatus, build_binding)
from karabogui.fonts import FONT_FAMILIES, get_font_size_from_dpi, get_qfont

from ..view import SceneView
from ..widget.container import ControllerContainer

NUM_TESTED_FONTS = 20
DEVICE_NAME = "Device"
PROPERTY_NAME = "availableScenes"
PROPERTY_PATH = f"{DEVICE_NAME}.{PROPERTY_NAME}"

GET_PROXY_PATH = "karabogui.sceneview.widget.container.get_proxy"

MAC_DPI_FACTOR = SCENE_DEFAULT_DPI / SCENE_MAC_DPI


class Object(Configurable):
    availableScenes = VectorString()


# fixtures
@pytest.fixture
def setup_widget_fonts(gui_app):
    view = SceneView()

    # Prepare a property proxy to create the controller widget
    # successfully
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    device = DeviceProxy(binding=binding,
                         server_id='Fake',
                         device_id=DEVICE_NAME,
                         status=ProxyStatus.OFFLINE)
    property_proxy = PropertyProxy(root_proxy=device,
                                   path=PROPERTY_NAME)
    property_proxy.value = ['bob', 'frank']
    yield view, property_proxy
    # The destroy will clear the cache
    view.destroy()


@pytest.fixture
def setup_scene_fonts(setup_widget_fonts):
    view, property_proxy = setup_widget_fonts
    # Prepare the fonts
    font_families = QFontDatabase().families()
    if len(font_families) > NUM_TESTED_FONTS:
        font_families = random.sample(font_families, NUM_TESTED_FONTS)
    yield view, property_proxy, font_families


@pytest.fixture
def setup_widget_fonts_macos(gui_app, mocker):
    # Create a QApplication with a mocked Mac OSX font size
    mocker.patch("karabogui.fonts.GUI_DPI_FACTOR", new=MAC_DPI_FACTOR)
    font_size = get_font_size_from_dpi(SCENE_FONT_SIZE)

    mocker.patch("karabogui.programs.base.SCENE_FONT_SIZE", new=font_size)
    # same setup as for normal test
    view = SceneView()
    schema = Object.getClassSchema()
    binding = build_binding(schema)
    device = DeviceProxy(binding=binding,
                         server_id='Fake',
                         device_id=DEVICE_NAME,
                         status=ProxyStatus.OFFLINE)
    property_proxy = PropertyProxy(root_proxy=device,
                                   path=PROPERTY_NAME)
    property_proxy.value = ['bob', 'frank']
    yield view, property_proxy
    # The destroy will clear the cache
    view.destroy()


# helper functions
def set_models_to_scene(view, property_proxy, models, mocker):
    mocker.patch(GET_PROXY_PATH, return_value=property_proxy)
    view.update_model(SceneModel(children=models))


def get_widget(view, model):
    return view._scene_obj_cache.get(model)


def assert_geometry(model, pos=(0, 0), size=(0, 0)):
    if isinstance(pos, QPoint):
        pos = (pos.x(), pos.y())
    if isinstance(size, QSize):
        size = (size.width(), size.height())

    assert model.x == pos[0]
    assert model.y == pos[1]
    assert model.width == size[0]
    assert model.height == size[1]


def assert_font_size(obj, expected=SCENE_FONT_SIZE):
    if isinstance(obj, ControllerContainer):
        qfont = obj.widget_controller.widget.font()
    elif isinstance(obj, QWidget):
        qfont = obj.font()
    else:
        # We want to compare the absolve value of the font size
        qfont = get_qfont(obj.font, adjust_size=False)
    assert qfont.pointSize() == expected


def assert_model_text_widgets(klass, view, property_proxy, mocker):
    # Trigger size hint calculation by not specifying model geometry
    model = klass(text="I am a long text qweqweqwe",
                  keys=[PROPERTY_PATH])
    font_size = get_font_size_from_dpi(SCENE_FONT_SIZE)
    # Add to scene
    set_models_to_scene(view, property_proxy, [model], mocker)
    widget = get_widget(view, model)
    assert_geometry(model, size=widget.size())
    assert_font_size(widget, expected=font_size)
    assert_font_size(model, expected=SCENE_FONT_SIZE)

    # Reload the model
    read_model = single_model_round_trip(model)
    assert_font_size(read_model, SCENE_FONT_SIZE)

    # Add the reloaded model to scene
    set_models_to_scene(view, property_proxy, [model], mocker)
    widget = get_widget(view, model)
    assert_geometry(read_model, pos=(model.x, model.y),
                    size=(model.width, model.height))
    assert_font_size(widget, expected=font_size)
    assert_font_size(read_model, expected=SCENE_FONT_SIZE)


def assert_model_scene_widgets(klass, view, property_proxy,
                               font_families, mocker):
    models = [
        klass(text=font, font=QFont(font).toString(), keys=[PROPERTY_PATH])
        for font in font_families]
    set_models_to_scene(view, property_proxy, models, mocker)

    for model in models:
        widget = view._scene_obj_cache[model]
        if isinstance(widget, ControllerContainer):
            # Get the internal widget from the controller container
            widget = widget.widget_controller.widget
        assert widget.font().family() in FONT_FAMILIES


# actual tests
def test_text_widgets(setup_widget_fonts, mocker):
    """Test the basic widgets fonts sizes"""
    view, property_proxy = setup_widget_fonts
    assert_model_text_widgets(LabelModel, view, property_proxy, mocker)
    assert_model_text_widgets(SceneLinkModel, view, property_proxy, mocker)
    assert_model_text_widgets(DeviceSceneLinkModel, view, property_proxy,
                              mocker)
    assert_model_text_widgets(StickerModel, view, property_proxy, mocker)
    assert_model_text_widgets(WebLinkModel, view, property_proxy, mocker)


def test_scene_text_widgets(setup_scene_fonts, mocker):
    """Test the scene font families replacement"""
    view, property_proxy, font_families = setup_scene_fonts
    assert_model_scene_widgets(LabelModel, view, property_proxy,
                               font_families, mocker)
    assert_model_scene_widgets(SceneLinkModel, view, property_proxy,
                               font_families, mocker)
    assert_model_scene_widgets(DeviceSceneLinkModel, view, property_proxy,
                               font_families, mocker)
    assert_model_scene_widgets(StickerModel, view, property_proxy,
                               font_families, mocker)
    assert_model_scene_widgets(WebLinkModel, view, property_proxy,
                               font_families, mocker)


@pytest.mark.skipif(system() == "Darwin",
                    reason="This test is mocking OSX from another OS.")
def test_text_widgets_on_macos(setup_widget_fonts_macos, mocker):
    """Test the macOS mock of the fonts"""
    view, property_proxy = setup_widget_fonts_macos
    mocker.patch("karabogui.fonts.GUI_DPI_FACTOR", new=MAC_DPI_FACTOR)
    assert_model_text_widgets(LabelModel, view, property_proxy, mocker)
    assert_model_text_widgets(SceneLinkModel, view, property_proxy, mocker)
    assert_model_text_widgets(DeviceSceneLinkModel, view, property_proxy,
                              mocker)
    assert_model_text_widgets(StickerModel, view, property_proxy, mocker)
    assert_model_text_widgets(WebLinkModel, view, property_proxy, mocker)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest

from karabogui.graph.common.api import ImageRegion
from karabogui.testing import GuiTestCase

from ..controller import (
    ProfileAggregator, ProfileAnalyzer, ProfileController, ProfilePlot)

X_LENGTH, Y_LENGTH = (5, 4)


class TestProfileAuxPlot(GuiTestCase):

    def setUp(self):
        super().setUp()
        self._controller = ProfileAggregator()
        self._controller.set_axes(x_data=np.arange(X_LENGTH),
                                  y_data=np.arange(Y_LENGTH))
        self._controller.on_trait_change(self._mock_slot, "stats")

    def tearDown(self):
        super().tearDown()
        self._controller.on_trait_change(self._mock_slot, "stats", remove=True)
        self._emitted_value = None

    def _mock_slot(self, value):
        self._emitted_value = value

    def test_basics(self):
        assert len(self._controller.controllers) == 2
        for aux_plot in self._controller.controllers.values():
            assert isinstance(aux_plot, ProfileController)
            assert isinstance(aux_plot.plot, ProfilePlot)
            assert isinstance(aux_plot.analyzer, ProfileAnalyzer)

        assert not self._controller.show_stats

    @pytest.mark.filterwarnings("ignore::scipy.optimize.OptimizeWarning")
    def test_analyze(self):
        # 1. Check default
        self._assert_analyze(valid=False, fitted=False)

        # 2. Check valid region
        self._analyze_region(valid=True)
        self._assert_analyze(valid=True, fitted=False)

        self._analyze_region(valid=True, fitted=True)
        self._assert_analyze(valid=True, fitted=True)

        # 3. Check invalid region
        self._analyze_region(valid=False)
        self._assert_analyze(valid=False, fitted=False)

        self._analyze_region(valid=False, fitted=True)
        self._assert_analyze(valid=False, fitted=False)

    def _analyze_region(self, valid=True, fitted=False):
        region = ImageRegion()
        if valid:
            image = np.arange(X_LENGTH * Y_LENGTH).reshape(Y_LENGTH, X_LENGTH)
            region = ImageRegion(region=image,
                                 region_type=ImageRegion.Area,
                                 x_slice=slice(X_LENGTH),
                                 y_slice=slice(Y_LENGTH))

        assert region.valid() is valid
        self._controller._enable_fitting(fitted)
        self._controller.process(region)

    def _assert_analyze(self, valid=True, fitted=True):
        for controller in self._controller.controllers.values():
            plot_item, analyzer = controller.plot, controller.analyzer
            data_item, fit_item = plot_item._data_item, plot_item._fit_item

            # Check profile values
            x_data, y_data = analyzer._x_profile, analyzer._y_profile
            if valid:
                data_assertion = self.assertArrayIsNotEmpty
                # Add one more vestigial data at the end because of step plot
                x_data = np.hstack((x_data, [x_data[-1] + 1]))
            else:
                data_assertion = self.assertArrayIsEmpty

            data_assertion(x_data)
            data_assertion(x_data)
            np.testing.assert_array_equal(data_item.xData, x_data)
            np.testing.assert_array_equal(data_item.yData, y_data)

            # Check fit values
            fit_assertion = (self.assertArrayIsNotEmpty if fitted
                             else self.assertArrayIsEmpty)
            x_fit, y_fit = analyzer._x_fit, analyzer._y_fit

            fit_assertion(x_fit)
            fit_assertion(y_fit)
            np.testing.assert_array_equal(fit_item.xData, x_fit)
            np.testing.assert_array_equal(fit_item.yData, y_fit)

    @pytest.mark.filterwarnings("ignore::scipy.optimize.OptimizeWarning")
    def test_enable_fitting(self):
        # 1. Check default
        self._assert_show_stats(valid=False)

        # 2. Check valid region
        self._analyze_region(valid=True)
        self._assert_show_stats(valid=True)

        # 3. Check invalid region
        self._analyze_region(valid=False)
        self._assert_show_stats(valid=False)

    def _assert_show_stats(self, valid=True):
        assertion = self.assertNotEqual if valid else self.assertEqual

        # Start from disabled stats
        if self._controller.show_stats:
            self._controller._enable_fitting(False)
            assert not self._controller.show_stats
            assert self._emitted_value is None

        # Show stats
        self._controller._enable_fitting(True)
        assert self._controller.show_stats
        assertion(self._emitted_value.html, '')

        # Hide stats
        self._controller._enable_fitting(False)
        assert not self._controller.show_stats
        assert self._emitted_value is None

    @staticmethod
    def assertArrayIsEmpty(array):
        np.testing.assert_array_equal(array, [])

    @staticmethod
    def assertArrayIsNotEmpty(array):
        np.testing.assert_raises(AssertionError,
                                 np.testing.assert_array_equal, array, [])



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on January 18, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import Counter
from copy import deepcopy

from karabo.common.api import ServerFlags
from karabo.native import Hash
from karabogui.singletons.api import get_topology


def _is_online(prefix, name):
    topology = get_topology()
    path = prefix + "." + name
    attrs = topology.get_attributes(path)
    return True if attrs else False


def is_server_online(serverId):
    """Check the system topology and return online server status"""
    return _is_online("server", serverId)


def is_device_online(deviceId):
    """Check the system topology and return online device status"""
    return _is_online("device", deviceId)


def is_macro_online(deviceId):
    """Check the system topology and return online macro status"""
    return _is_online("macro", deviceId)


def get_macro_servers(development=False):
    """ get an ordered list of macro servers

    Walk the system topology and return a list of
    macro servers ordered from least to most loaded

    :param development: Denote if development macro servers should be included
                        The default is `False`
    """
    topology = get_topology()
    # macro_server is a Counter object that will contain one entry per macro
    # server set to the number of macros in that macro server + 1
    macro_servers = Counter()
    # macro_servers_set contains all server of `macro` type.
    macro_servers_set = set()

    def visitor(node):
        nonlocal macro_servers
        attrs = node.attributes
        if attrs.get("type") == "server" and attrs.get("lang") == "macro":
            # Exclude development servers
            if (not development and attrs.get("serverFlags", 0)
                    & ServerFlags.Development == ServerFlags.Development):
                return
            macro_servers[attrs["serverId"]] += 1
            macro_servers_set.add(node.node_id)
        elif (attrs.get("type") == "macro"
              and attrs.get("serverId") is not None
              and attrs.get("serverId") != "__none__"):
            macro_servers[attrs["serverId"]] += 1

    topology.visit_system_tree(visitor)

    return [server for server in sorted(macro_servers, key=macro_servers.get)
            if server in macro_servers_set]


def getTopology():
    """Return a deepcopy of the system topology hash"""
    topology = get_topology()._system_hash or Hash()
    return deepcopy(topology)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from traits.api import (
    Array, CArray, Dict, Either, Enum, Event, HasStrictTraits, Instance, List,
    Property, String, Trait, TraitError, TraitHandler, Undefined,
    cached_property)

from karabo.common import const
from karabo.common.states import State
from karabo.native import (
    AccessLevel, AccessMode, Assignment, Hash, HashList, Timestamp)

from .trait_types import CBool, Float, NumpyRange

KEY_DISPLAYED_NAME = const.KARABO_SCHEMA_DISPLAYED_NAME
KEY_DISPLAY_TYPE = const.KARABO_SCHEMA_DISPLAY_TYPE
KEY_ACCMODE = const.KARABO_SCHEMA_ACCESS_MODE
KEY_ASSIGNMENT = const.KARABO_SCHEMA_ASSIGNMENT
KEY_ACCLVL = const.KARABO_SCHEMA_REQUIRED_ACCESS_LEVEL
KEY_OPT = const.KARABO_SCHEMA_OPTIONS
KEY_UNIT = const.KARABO_SCHEMA_UNIT_SYMBOL
KEY_UNITPREFIX = const.KARABO_SCHEMA_METRIC_PREFIX_SYMBOL
KEY_ROW_SCHEMA = const.KARABO_SCHEMA_ROW_SCHEMA


class BaseBinding(HasStrictTraits):
    """The base class for all bindings. It represents a single node in an
    object created from a schema. It has a dictionary of attributes and a
    `value` trait which contains the value of the node. The value is validated
    using normal ``Traits`` validation.

    :param attributes: The dictionary of the attributes belonging to the
                       `BaseBinding`.
    :param config_update: ``Traits.Event`` which fires when the value is
                          updated externally
    :param historic_data: ``Traits.Event`` which fires when historic data
                          arrives for this object node. The data is contained
                          in the new value passed to notification handlers
    :param value: The value contained in this node. Derived classes should
                  redefine this. The default is ``Traits.Undefined``.
    :param timestamp: The timestamp when the value was last set on the device.
                      Instance of `karabo.native.Timestamp`.


    Attribute shortcuts for often used properties
    ---------------------------------------------

    :param displayedName: The displayed name of the property
    :param displayType: The display type of the property. This might provide
                         more options later on.
    :param accessMode: The access mode specification of the binding taking an
                        enum of `karabo.native.AccessMode`.
    :param assignment: The assignment setting with an enum of
                       `karabo.native.Assignment`.
    :param options: The list of options for this base binding.
    :param requiredAccessLevel: The required access level with enums of
                                  `karabo.native.AccessLevel`.
    :param unit_label: The unit label string combining the unit and metric
                       prefix.
    """
    # Attributes property copied from the object schema, we control the
    # setter of the attributes so their shortcuts can be updated in time
    # but avoid hammering the configurator

    attributes = Property
    _attributes = Dict
    # When the value was last set on the device
    timestamp = Instance(Timestamp)
    value = Undefined
    config_update = Event
    historic_data = Event

    # Attribute shortcuts
    displayedName = String
    displayType = String
    accessMode = Enum(*AccessMode)
    assignment = Enum(*Assignment)
    options = List
    requiredAccessLevel = Enum(*AccessLevel)
    unit_label = String

    def is_allowed(self, state):
        """Return True if the given `state` is an allowed state for this
        binding.

        :param state: Ideally an instance of `karabo.native.State` otherwise
                      a string that can be casted to a ``State`` enum.
        """
        if isinstance(state, State):
            state = state.value

        alloweds = self._attributes.get(const.KARABO_SCHEMA_ALLOWED_STATES, [])
        return alloweds == [] or state in alloweds

    def _get_attributes(self):
        return self._attributes

    def _set_attributes(self, value):
        self._attributes = value

    def update_attributes(self, attrs):
        """Silently update attributes dictionary"""
        if attrs:
            traits = {'_attributes': self._attributes}
            traits['_attributes'].update(attrs)
            self.trait_setq(**traits)
            self._update_shortcuts(attrs)

    def _displayedName_default(self):
        return self._attributes.get(KEY_DISPLAYED_NAME, '')

    def _displayType_default(self):
        return self._attributes.get(KEY_DISPLAY_TYPE, '')

    def _accessMode_default(self):
        mode = self._attributes.get(KEY_ACCMODE)
        return AccessMode.UNDEFINED if mode is None else AccessMode(mode)

    def _assignment_default(self):
        assign = self._attributes.get(KEY_ASSIGNMENT)
        return Assignment.OPTIONAL if assign is None else Assignment(assign)

    def _options_default(self):
        return self._attributes.get(KEY_OPT, [])

    def _requiredAccessLevel_default(self):
        level = self._attributes.get(KEY_ACCLVL)
        return AccessLevel.OBSERVER if level is None else AccessLevel(level)

    def _unit_label_default(self):
        attrs = self._attributes
        unit = attrs.get(KEY_UNITPREFIX, '') + attrs.get(KEY_UNIT, '')
        return unit

    def _update_shortcuts(self, attrs):
        if KEY_DISPLAYED_NAME in attrs:
            self.displayedName = attrs[KEY_DISPLAYED_NAME]
        if KEY_DISPLAY_TYPE in attrs:
            self.displayType = attrs[KEY_DISPLAY_TYPE]
        if KEY_ACCMODE in attrs:
            self.accessMode = AccessMode(attrs[KEY_ACCMODE])
        if KEY_ACCLVL in attrs:
            self.requiredAccessLevel = AccessLevel(attrs[KEY_ACCLVL])
        if KEY_ASSIGNMENT in attrs:
            self.assignment = Assignment(attrs[KEY_ASSIGNMENT])
        if KEY_OPT in attrs:
            self.options = attrs[KEY_OPT]
        if KEY_UNIT in attrs or KEY_UNITPREFIX in attrs:
            unit = attrs.get(KEY_UNITPREFIX, '') + attrs.get(KEY_UNIT, '')
            self.unit_label = unit

    def check(self, value):
        """Check and validate the value of this binding"""
        options = self._attributes.get(KEY_OPT, None)
        if options is not None and value not in options:
            raise TraitError(f"The value {value} is not in the allowed options"
                             f"{options}")

        return self.validate_trait("value", value)


class BindingNamespace:
    """A namespace which is iterable and remembers the order in which values
    were added. Additionally, the type of the values in the namespace can be
    restricted with an `item_type` which is a type or tuple of types.
    """

    def __init__(self, *, item_type=object):
        priv_prefix = '_' + type(self).__name__
        # Avoid __setattr__ recursion by modifying __dict__ directly
        self.__dict__[priv_prefix + '__type'] = item_type
        self.__dict__[priv_prefix + '__names'] = []
        self.__dict__[priv_prefix + '__priv_prefix'] = priv_prefix

    def clear_namespace(self):
        """Remove all items from the namespace"""
        func = super().__delattr__
        for key in self.__names:
            func(key)
        self.__names.clear()

    def __contains__(self, key):
        return key in self.__names

    def __delattr__(self, key):
        if not key.startswith(self.__priv_prefix):
            super().__delattr__(key)
            self.__names.remove(key)

    def __setattr__(self, key, value):
        if not isinstance(value, self.__type):
            msg = 'Attributes of type {} are expected, but type {} was passed'
            raise ValueError(msg.format(self.__type, type(value)))

        super().__setattr__(key, value)
        if key not in self.__names and not key.startswith(self.__priv_prefix):
            self.__names.append(key)

    def __iter__(self):
        return iter(self.__names)

    def __len__(self):
        return len(self.__names)

    def __repr__(self):
        names = '\n'.join('\t' + n + ': ' + repr(getattr(self, n))
                          for n in self)
        return f'{{\n{names}\n}}'


class BindingRoot(BaseBinding):
    """The root node of an object binding which was created from a schema.
    """
    # The top-level namespace of the object
    value = Instance(BindingNamespace, kw={'item_type': BaseBinding})
    # The name of the class represented
    class_id = String
    # An event which fires when the schema changes
    schema_update = Event
    # A cache for visible children names {accesslevel: [names]}
    # All bindings with their value type as BindingNamespace should have
    # this trait to cache the names in the value according to access level
    children_names = Dict()

    def _schema_update_fired(self):
        # when schema changes, clear cached children names
        self.children_names = {}


# =============================================================================
# Binding classes for each type expected in Karabo (see Schema.hh and Types.hh)
#

class BoolBinding(BaseBinding):
    value = CBool
    # number type binding (Bool, Float and Int) could have options in type of
    # numpy array, CArray will cast list to a numpy array. Be aware that when
    # comparing numpy arrays, result is an array as well.
    options = CArray

    def _options_default(self):
        return self._attributes.get(KEY_OPT, [])


class _ByteArrayHandler(TraitHandler):
    """A trait handler to work around middlelater stupidity.
    """

    def validate(self, obj, name, value):
        if isinstance(value, bytearray):
            return value
        elif isinstance(value, bytes):
            return bytearray(value)
        self.error(obj, name, value)


class ByteArrayBinding(BaseBinding):
    value = Trait(_ByteArrayHandler())


class CharBinding(BaseBinding):
    value = String(maxlen=1)


class FloatBinding(BaseBinding):
    value = Float
    options = CArray

    def _options_default(self):
        return self._attributes.get(KEY_OPT, [])

    def check(self, value):
        value = super().check(value)
        # Karabo attribute check
        low, high = self.getMinMax()
        if value < low:
            raise TraitError(f"The value {value} is lower than {low}")
        if value > high:
            raise TraitError(f"The value {value} is higher than {high}")

        return value

    def getMinMax(self):
        attrs = self._attributes
        value_type = attrs.get(const.KARABO_SCHEMA_VALUE_TYPE)
        if value_type in ('FLOAT', 'COMPLEX_FLOAT'):
            info = np.finfo(np.float32)
        else:
            info = np.finfo(np.float64)

        low = attrs.get(const.KARABO_SCHEMA_MIN_EXC)
        if low is not None:
            low = low * (1 + np.sign(low) * info.eps) + info.tiny
        else:
            low = attrs.get(const.KARABO_SCHEMA_MIN_INC, info.min)

        high = attrs.get(const.KARABO_SCHEMA_MAX_EXC)
        if high is not None:
            high = high * (1 - np.sign(high) * info.eps) - info.tiny
        else:
            high = attrs.get(const.KARABO_SCHEMA_MAX_INC, info.max)

        return low, high


class HashBinding(BaseBinding):
    value = Instance(Hash)


class IntBinding(BaseBinding):
    """The base class for all integer binding types"""
    options = CArray

    def _options_default(self):
        return self._attributes.get(KEY_OPT, [])

    def check(self, value):
        value = super().check(value)
        # Karabo attribute check
        low, high = self.getMinMax()
        if value < low:
            raise TraitError(f"The value {value} is lower than {low}")
        if value > high:
            raise TraitError(f"The value {value} is higher than {high}")

        return value

    def getMinMax(self):
        range_trait = self.trait('value').handler
        value_range = range_trait._low, range_trait._high
        if range_trait._exclude_low:
            value_range = (value_range[0] + 1, value_range[1])
        if range_trait._exclude_high:
            value_range = (value_range[0], value_range[1] - 1)

        attrs = self._attributes
        low = attrs.get(const.KARABO_SCHEMA_MIN_EXC)
        if low is not None:
            low += 1
        else:
            low = attrs.get(const.KARABO_SCHEMA_MIN_INC, value_range[0])

        high = attrs.get(const.KARABO_SCHEMA_MAX_EXC)
        if high is not None:
            high -= 1
        else:
            high = attrs.get(const.KARABO_SCHEMA_MAX_INC, value_range[1])

        return low, high


class SignedIntBinding(IntBinding):
    """The base class for all signed integer binding types"""


class UnsignedIntBinding(IntBinding):
    """The base class for all unsigned integer binding types"""


class Int8Binding(SignedIntBinding):
    value = NumpyRange(low=(-1 - (1 << 7)), high=(1 << 7), value=0,
                       exclude_low=True, exclude_high=True, dtype=np.int8)


class Int16Binding(SignedIntBinding):
    value = NumpyRange(low=(-1 - (1 << 15)), high=(1 << 15), value=0,
                       exclude_low=True, exclude_high=True, dtype=np.int16)


class Int32Binding(SignedIntBinding):
    value = NumpyRange(low=(-1 - (1 << 31)), high=(1 << 31), value=0,
                       exclude_low=True, exclude_high=True, dtype=np.int32)


class Int64Binding(SignedIntBinding):
    value = NumpyRange(low=(-1 - (1 << 63)), high=(1 << 63), value=0,
                       exclude_low=True, exclude_high=True, dtype=np.int64)


class NodeBinding(BaseBinding):
    value = Instance(BindingNamespace)
    # Visible children names {accesslevel: [names]}
    children_names = Dict()


class ImageBinding(NodeBinding):
    pass  # Nothing to add. We just need a different class for Image


class NDArrayBinding(NodeBinding):
    pass  # Nothing to add. We just need a different class for NDArray


class WidgetNodeBinding(NodeBinding):
    pass  # Nothing to add. We just need a different class for WidgetNode


class NoneBinding(BaseBinding):
    value = Undefined


class PipelineOutputBinding(NodeBinding):
    pass  # Nothing to add. We just need a different class for PipelineOutput


class SchemaBinding(BaseBinding):
    value = Instance(object)


class SlotBinding(NodeBinding):
    pass  # Nothing to add. We just need a different class for Slots


class StringBinding(BaseBinding):
    value = String


class Uint8Binding(UnsignedIntBinding):
    value = NumpyRange(low=0, high=(1 << 8), value=0, exclude_high=True,
                       dtype=np.uint8)


class Uint16Binding(UnsignedIntBinding):
    value = NumpyRange(low=0, high=(1 << 16), value=0, exclude_high=True,
                       dtype=np.uint16)


class Uint32Binding(UnsignedIntBinding):
    value = NumpyRange(low=0, high=(1 << 32), value=0, exclude_high=True,
                       dtype=np.uint32)


class Uint64Binding(UnsignedIntBinding):
    value = NumpyRange(low=0, high=(1 << 64), value=0, exclude_high=True,
                       dtype=np.uint64)


class VectorBinding(BaseBinding):
    """The base class for all vector binding types"""

    def check(self, value):
        value = super().check(value)
        # Karabo attribute check for minSize and maxSize
        attributes = self._attributes
        minSize = attributes.get(const.KARABO_SCHEMA_MIN_SIZE)
        if minSize is not None and len(value) < minSize:
            raise TraitError(f"Vector with size {len(value)} is shorter than "
                             f"the allowed size of {minSize}")
        maxSize = attributes.get(const.KARABO_SCHEMA_MAX_SIZE)
        if maxSize is not None and len(value) > maxSize:
            raise TraitError(f"Vector with size {len(value)} is larger than "
                             f"the allowed size of {maxSize}")
        return value


class VectorNumberBinding(VectorBinding):
    """The base class for all vector binding types which contain types
    supported by numpy.
    """


class VectorBoolBinding(VectorNumberBinding):
    value = Array(dtype='bool')


class _ByteHandler(TraitHandler):
    """A trait handler to handle string to bytes conversion
    """

    def validate(self, obj, name, value):
        if isinstance(value, bytes):
            return value
        elif isinstance(value, str):
            return value.encode()
        self.error(obj, name, value)


class VectorCharBinding(VectorBinding):
    value = Trait(_ByteHandler())


class VectorDoubleBinding(VectorNumberBinding):
    value = Array(dtype='float64', shape=(None,))


class VectorFloatBinding(VectorNumberBinding):
    value = Array(dtype='float32', shape=(None,))


class VectorHashBinding(VectorBinding):
    """The VectorHash Binding for Table elements of devices

    :param row_schema: Instance of Hash containing the row schema
    :param bindings: A dictionary of bindings associated to the column keys
    """
    value = Either((Instance(HashList), List))
    row_schema = Instance(Hash)
    bindings = Property(depends_on="row_schema")

    @cached_property
    def _get_bindings(self):
        from .builder import _BINDING_MAP
        bindings = {}
        for key in self.row_schema.getKeys():
            attrs = self.row_schema[key, ...]
            value_type = attrs[const.KARABO_SCHEMA_VALUE_TYPE]
            binding_factory = _BINDING_MAP[value_type]
            bindings[key] = binding_factory(attributes=attrs, value=Undefined)
        return bindings

    def get_binding(self, key):
        return self.bindings.get(key)

    def _row_schema_default(self):
        row_schema = self._attributes.get(KEY_ROW_SCHEMA)
        if row_schema is None:
            return Hash()

        return row_schema.hash

    def _update_shortcuts(self, attrs):
        super()._update_shortcuts(attrs)
        row_schema = attrs.get(KEY_ROW_SCHEMA)
        if row_schema is None:
            return Hash()

        self.row_schema = row_schema.hash


class VectorInt8Binding(VectorNumberBinding):
    value = Array(dtype='int8', shape=(None,))


class VectorInt16Binding(VectorNumberBinding):
    value = Array(dtype='int16', shape=(None,))


class VectorInt32Binding(VectorNumberBinding):
    value = Array(dtype='int32', shape=(None,))


class VectorInt64Binding(VectorNumberBinding):
    value = Array(dtype='int64', shape=(None,))


class VectorNoneBinding(VectorBinding):
    value = List(Undefined)


class VectorStringBinding(VectorBinding):
    value = List(String)


class VectorUint8Binding(VectorNumberBinding):
    value = Array(dtype='uint8', shape=(None,))


class VectorUint16Binding(VectorNumberBinding):
    value = Array(dtype='uint16', shape=(None,))


class VectorUint32Binding(VectorNumberBinding):
    value = Array(dtype='uint32', shape=(None,))


class VectorUint64Binding(VectorNumberBinding):
    value = Array(dtype='uint64', shape=(None,))


class TableBinding(VectorHashBinding):
    pass  # Nothing to add. We just need a different class for Tables



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from collections import defaultdict
from collections.abc import Iterable
from itertools import chain

from traits.api import Constant

from karabo.common.scenemodel.api import BaseWidgetObjectData

from .base import BaseBindingController
from .util import get_class_const_trait

# Module global registry
_controller_registry = defaultdict(set)
_controller_models = defaultdict(set)


def get_compatible_controllers(binding_instance, can_edit=False):
    """Returns a list of `BaseBindingController` subclasses which are capable
    of creating a view for the given `BaseBinding` object instance.
    """
    def check_compatibility(klass):
        if can_edit == get_class_const_trait(klass, '_can_edit'):
            checker = get_class_const_trait(klass, '_is_compatible')
            return checker(binding_instance)
        return False

    key = type(binding_instance)
    # Sort based on priority
    klasses = sorted(_controller_registry[key],
                     key=lambda x: get_class_const_trait(x, '_priority'),
                     reverse=True)
    # Give the classes a chance to reject
    klasses = [klass for klass in klasses if check_compatibility(klass)]
    return klasses


def get_model_controller(scene_model):
    """Given a scene model instance, return the appropriate controller class
    to use in the scene view.
    """
    def _with_attribute(controllers, attr_name, attr_value):
        for klass in controllers:
            if attr_value == get_class_const_trait(klass, attr_name):
                return klass

    will_edit = scene_model.parent_component != 'DisplayComponent'
    controller_klass = getattr(scene_model, 'klass', '')
    klasses = _controller_models.get(type(scene_model), [])

    if controller_klass != '':
        # If a class name is specified by the model, use that
        return _with_attribute(klasses, '_klassname', controller_klass)
    elif len(klasses) > 1:
        # Use the editability to differentiate
        return _with_attribute(klasses, '_can_edit', will_edit)
    elif klasses:
        # We have to return the only controller that was found.
        # `klasses` might be a `set` and sets don't allow indexing
        return next(iter(klasses))
    else:
        return None


def get_scene_model_class(klass):
    """Return the scene model class for a given `BaseBindingController`
    subclass which has been registered.
    """
    trait = klass.class_traits()['model']
    instance_trait = trait.trait_type

    # Controller classes MUST define their scene model class!
    assert instance_trait.klass is not BaseWidgetObjectData
    return instance_trait.klass


class register_binding_controller:
    """A class decorator for `BaseBindingController` subclasses which registers
    the widget for the various UI builders and adds some useful metadata.

    :param ui_name: The name seen by users in the widget selection menu
    :param klassname: The name of the widget in the scene file format. This is
                      used for disambiguation in some of the legacy widgets.
    :param binding_type: A tuple of `BaseBinding` classes which denotes the
                         types understood by the controller.
    :param can_edit: If True, this controller is recognized as an editor.
    :param is_compatible: A callable which gives the controller extra control
                          over the process of checking a binding instance for
                          compatibility with this controller. The callable
                          should accept one argument of type `BaseBinding`.
    :param priority: A priority value for this controller which can influence
                     sorting and picking of defaults when dragging from the
                     configurator.
    :param can_show_nothing: If True, the controller should be able to tolerate
                             calls of its `value_update` and `state_update`
                             with binding instances whose timestamp is None.
                             see `BaseBindingController.finish_initialization`
    """
    def __init__(self, *, ui_name='', klassname='', binding_type=(),
                 is_compatible=None, priority=0, can_edit=False,
                 can_show_nothing=True):
        assert klassname != '', '`klassname` must be provided!'

        # Expand all binding types into a set of all possible subclasses
        if not isinstance(binding_type, Iterable):
            binding_type = set(_expand_binding(binding_type))
        else:
            accum = [_expand_binding(t) for t in binding_type]
            binding_type = set(chain(*accum))

        self.ui_name = str(ui_name)
        self.klassname = str(klassname)
        self.binding_type = tuple(binding_type)
        self.is_compatible = is_compatible or (lambda binding: True)
        self.priority = priority
        self.can_edit = bool(can_edit)
        self.can_show_nothing = bool(can_show_nothing)

    def __call__(self, klass):
        assert issubclass(klass, BaseBindingController)

        binding_type_trait = Constant(self.binding_type)
        klass.add_class_trait('_binding_type', binding_type_trait)
        klass.add_class_trait('_ui_name', Constant(self.ui_name))
        klass.add_class_trait('_klassname', Constant(self.klassname))
        klass.add_class_trait('_is_compatible', Constant(self.is_compatible))
        klass.add_class_trait('_priority', Constant(self.priority))
        klass.add_class_trait('_can_edit', Constant(self.can_edit))
        klass.add_class_trait('_can_show_nothing',
                              Constant(self.can_show_nothing))

        # Register `klass` for building UIs
        model_klass = get_scene_model_class(klass)
        _controller_models[model_klass].add(klass)
        for t in self.binding_type:
            _controller_registry[t].add(klass)
        return klass


# --------------------------------------------------------------------------
# Private things

def _expand_binding(binding_klass):
    """Given a binding class, generate all the classes descended from that
    class.
    """
    yield binding_klass
    for klass in binding_klass.__subclasses__():
        yield from _expand_binding(klass)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from copy import deepcopy

import numpy as np
from pyqtgraph import ColorMap, PlotItem
from qtpy.QtCore import QRectF, Signal, Slot
from qtpy.QtGui import QFont

from karabogui.graph.common.api import (
    COLORMAPS, ArrayExporter, AspectRatio, ExportTool, ImageExporter,
    create_axis_items)
from karabogui.graph.common.const import (
    AXIS_ITEMS, AXIS_X, AXIS_Y, DEFAULT_LABEL_X, DEFAULT_LABEL_Y,
    DEFAULT_OFFSET_X, DEFAULT_OFFSET_Y, DEFAULT_SCALE_X, DEFAULT_SCALE_Y,
    DEFAULT_UNITS_X, DEFAULT_UNITS_Y, ROTATION_FACTOR, TF_FLIPPED, TF_ROTATION,
    TF_SCALING, TF_TRANSLATION)

from .item import KaraboImageItem
from .viewbox import KaraboImageViewBox


def axis_tick_font():
    font = QFont("Source Sans Pro")
    font.setPixelSize(10)
    return font


TICK_AXIS = ["top", "left"]
AXIS_ORDER = "row-major"


class KaraboImagePlot(PlotItem):
    imageTransformed = Signal()
    imageLevelsChanged = Signal(object)
    imageAxesChanged = Signal()

    def __init__(self, parent=None):
        super().__init__(
            viewBox=KaraboImageViewBox(parent=None),
            axisItems=create_axis_items(axes_with_ticks=TICK_AXIS),
            enableMenu=False,
            parent=parent)

        # Initialize widgets
        self.imageItem = KaraboImageItem(parent=self)
        # aligning the image coords to view coords
        self.imageItem.axisOrder = AXIS_ORDER

        self.addItem(self.imageItem)

        # Improve axes rendering
        for axis in AXIS_ITEMS:
            axis_item = self.getAxis(axis)
            axis_item.setZValue(0)

        self.show_all_axis()

        # Initialize values
        self._axes = [AXIS_X, AXIS_Y]
        self._image_set = False
        self._transform = {}
        self._units = []
        self.aspect_ratio = AspectRatio.PixelDependent

        self._base_transform = {}
        self._base_units = []

        # Container for ranges
        self._axes_data = [np.array([]), np.array([])]
        self.transformed_axes = None

        # Plot items
        self.printer = None

        x_label = {"units": DEFAULT_UNITS_X, "label": DEFAULT_LABEL_X}
        y_label = {"units": DEFAULT_UNITS_Y, "label": DEFAULT_LABEL_Y}
        self._labels = [x_label, y_label]

        # Set default values for the transform
        self._set_default_transform()
        self._flip()

        # Connect signals
        self.vb.sigResized.connect(self.imageItem.reset_downsampling)

    # ---------------------------------------------------------------------
    # PyQt slots

    @Slot()
    def _restore_view(self):
        """Restores the image item to the original transform, which is
        usually no scaling, translation, and rotation.
        """
        # Restore original transform
        self._revert_transform()

        # Only update the image after all operation
        self._apply_transform(update=False)

        # Flip axes if needed
        self._flip()

        self.vb.enableAutoRange()

    # ---------------------------------------------------------------------
    # Properties

    @property
    def axes_transform(self):
        return self._transform

    @property
    def axes_labels(self):
        return deepcopy(self._labels)

    @property
    def image(self):
        return self.imageItem.image

    @property
    def image_set(self):
        return self._image_set

    # ---------------------------------------------------------------------
    # Public methods

    def set_aspect_ratio(self, ratio=AspectRatio.NoAspectRatio):
        self.aspect_ratio = ratio
        if self.aspect_ratio == AspectRatio.NoAspectRatio:
            self.vb.setAspectLocked(False)
        elif self.aspect_ratio == AspectRatio.PixelDependent:
            x_scale, y_scale = self._transform[TF_SCALING]
            self.vb.setAspectLocked(True, ratio=y_scale / x_scale)
        elif self.aspect_ratio == AspectRatio.ScaleDependent:
            self.vb.setAspectLocked(True, ratio=1)

    def setData(self, data, update=True):
        """Convenience wrapper for a generic interface with plot widgets"""
        self.set_image(data, update)

    def set_image(self, image, update=True):
        if image is None or image.size == 0:
            return

        if not self._image_set:
            self._image_set = True

        should_update = False
        if self._axes_changed(image):
            # Axes are different, need to remap them
            self._axes_data = [np.arange(image.shape[1]),
                               np.arange(image.shape[0])]
            self._update_axes_transforms()
            should_update = True

        self.imageItem.setImage(image)
        if should_update and update:
            self._restore_view()

    def set_transform(self, x_scale, y_scale, x_translate, y_translate,
                      aspect_ratio=AspectRatio.PixelDependent, default=False):
        scaling = np.array([x_scale, y_scale], dtype=np.float64)
        translation = np.array([x_translate, y_translate], dtype=np.float64)

        if default:
            self._base_transform[TF_SCALING] = scaling
            self._base_transform[TF_TRANSLATION] = translation
            self._revert_transform()
        else:
            self._transform[TF_SCALING] = scaling
            self._transform[TF_TRANSLATION] = translation

        self.aspect_ratio = aspect_ratio

        if not default:
            self._apply_transform()

    def set_translation(self, x_translate=None, y_translate=None, update=True):
        """Public method to translate the image. Used by extensions"""
        if x_translate is None and y_translate is None:
            return

        if x_translate is not None:
            self._transform[TF_TRANSLATION][0] = x_translate
        if y_translate is not None:
            self._transform[TF_TRANSLATION][1] = y_translate

        self._apply_transform(update)

    def set_colormap(self, cmap):
        """Sets the colormap of the image by setting the image item LUT.
            Also updates the information of the selected pixel.

        :param str cmap:
            Colormap name. Should be among COLORMAPS.
        """
        lut = ColorMap(*zip(*COLORMAPS[cmap])).getLookupTable(
            alpha=False)
        self.imageItem.setLookUpTable(lut)

    @Slot(object)
    def set_image_levels(self, levels):
        """Sets the image levels. Also updates the information of the
        selected pixel.

        :param object levels:
             The image levels. Can be [float, float] or None
        """
        self.imageItem.auto_levels = levels is None
        self.imageItem.setLevels(levels)
        self.imageLevelsChanged.emit(levels)

    def hide_all_axis(self):
        for v in AXIS_ITEMS:
            self.hideAxis(v)

    def show_all_axis(self):
        for axis in AXIS_ITEMS:
            self.showAxis(axis)

    def rotate_axis(self, theta):
        raise NotImplementedError()

    def flip_axis(self, axis):
        raise NotImplementedError()

    def set_viewbox_menu(self, menu=None):
        enabled = menu is not None
        if menu is not None:
            self.vb.menu = menu
        self.vb.setMenuEnabled(enabled)

    def set_range(self, x_range, y_range, padding=0):
        self.vb.setXRange(*x_range, padding=padding)
        self.vb.setYRange(*y_range, padding=padding)

    def set_label(self, axis=0, text='', units=''):
        formatted_units = f" ({units})" if units else units

        show = True if (text != "" or units != "") else False

        for ax in self._axes[axis]:
            # set AxisItem labels
            self.getAxis(ax).setLabel(text + formatted_units)
            # XXX: We circumvent a PyQtGraph bug here and set the visibility
            # ourselves! Setting a label and removing it, gives PyQtGraph
            # problems.
            self.showLabel(ax, show=show)

        labels = self._labels[axis]
        labels["label"] = text
        labels["units"] = units

    @Slot(ExportTool)
    def export(self, export_type):
        """Exports the image according to the desired format"""
        if export_type == ExportTool.Image:
            exporter = ImageExporter(self.scene())
        elif export_type == ExportTool.Data:
            exporter = ArrayExporter(self.imageItem.image)
        else:
            raise LookupError(f"No exporter found for {export_type}")

        exporter.export()

    def _set_default_transform(self):
        self._base_transform[TF_SCALING] = np.array(
            [DEFAULT_SCALE_X, DEFAULT_SCALE_Y], dtype=np.float64)
        self._base_transform[TF_TRANSLATION] = np.array(
            [DEFAULT_OFFSET_X, DEFAULT_OFFSET_Y], dtype=np.float64)
        self._base_transform[TF_ROTATION] = 0
        # x-axis is not flipped, while y-axis is. This is due to the axis order
        self._base_transform[TF_FLIPPED] = [False, True]

        self._revert_transform()

    def _revert_transform(self):
        self._transform = deepcopy(self._base_transform)
        self._units = deepcopy(self._base_units)

    # ---------------------------------------------------------------------
    # Transform methods

    def _axes_changed(self, image):
        """Returns True/False if the axes information changed in relation to
        an image"""
        size_x = self._axes_data[0].size
        size_y = self._axes_data[1].size
        new_size_y, new_size_x = image.shape[:2]

        return (size_x != new_size_x) or (size_y != new_size_y)

    def _update_axes_transforms(self):
        """Updates the transformed axes"""
        translation = self._transform[TF_TRANSLATION]
        scaling = self._transform[TF_SCALING]

        self.transformed_axes = []
        for i, axis in enumerate(self._axes_data):
            self.transformed_axes.append((axis * scaling[i]) + translation[i])

        self.imageAxesChanged.emit()

    def _apply_transform(self, update=True):
        """Applies transformation on the image item such as scale, translate,
           and rotation. Other plot items such as ROIs and indicators are
           transformed as well."""
        rotation = self._transform[TF_ROTATION]
        translation = self._transform[TF_TRANSLATION]
        scaling = self._transform[TF_SCALING]

        self.imageItem.set_transform(
            scaling=scaling * ROTATION_FACTOR[rotation],
            translation=translation / scaling,
            rotation=rotation)

        self._update_axes_transforms()

        # calculate new image aspect ratio
        if self.aspect_ratio == AspectRatio.NoAspectRatio:
            self.vb.setAspectLocked(False)
        elif self.aspect_ratio == AspectRatio.PixelDependent:
            x_scale, y_scale = scaling
            self.vb.setAspectLocked(True, ratio=y_scale / x_scale)
        elif self.aspect_ratio == AspectRatio.ScaleDependent:
            self.vb.setAspectLocked(True, ratio=1)

        self.imageTransformed.emit()
        if update:
            self.vb.enableAutoRange()

    def mapRectFromTransform(self, rect):
        translation = self._transform[TF_TRANSLATION]
        scaling = self._transform[TF_SCALING]

        pos = np.array([rect.x(), rect.y()])
        abs_pos = ((pos - translation) / scaling).astype(int)

        size = np.array([rect.width(), rect.height()])
        abs_size = np.round(size / scaling)

        return QRectF(*(tuple(abs_pos) + tuple(abs_size)))

    def mapRectToTransform(self, rect):
        translation = self._transform[TF_TRANSLATION]
        scaling = self._transform[TF_SCALING]

        pos = np.array([rect.x(), rect.y()])
        trans_pos = pos * scaling + translation

        size = np.array([rect.width(), rect.height()])
        trans_size = size * scaling

        return QRectF(*(tuple(trans_pos) + tuple(trans_size)))

    def _flip(self):
        """Flips the image by inverting the viewbox
           and adjusting the shown axes"""
        flipped_x = self._transform[TF_FLIPPED][0]
        self.vb.invertX(flipped_x)

        flipped_y = self._transform[TF_FLIPPED][1]
        self.vb.invertY(flipped_y)

        major_x = "right" if flipped_x else "left"
        major_y = "top" if flipped_y else "bottom"
        major_axes = [major_x, major_y]

        for axis in AXIS_ITEMS:
            axis_item = self.getAxis(axis)
            show_values = axis in major_axes
            axis_item.setStyle(**{"showValues": show_values})

    # ---------------------------------------------------------------------
    # Reimplemented `PyQtGraph` methods

    def setMenuEnabled(self, enableMenu=True, enableViewBoxMenu=None):
        """Reimplemented function to circumvent behavior of pyqtgraph

        Note: This patch is not required in pyqtgraph > 0.11.1
        """
        self._menuEnabled = enableMenu



from copy import deepcopy

import pytest

from karabogui.binding.api import (
    DeviceProxy, ProxyStatus, apply_default_configuration, build_binding)
from karabogui.dialogs.compare_device_configurations import (
    CompareDeviceConfigurationsDialog, DeviceSelectorDialog)
from karabogui.testing import (
    get_device_schema, get_device_schema_allowed_state, singletons,
    system_hash)
from karabogui.topology.api import SystemTopology


@pytest.fixture()
def dialog(gui_app):
    binding1 = build_binding(get_device_schema_allowed_state())
    proxy1 = DeviceProxy(binding=binding1, server_id="Test",
                         status=ProxyStatus.ONLINE)
    apply_default_configuration(proxy1.binding)

    binding2 = build_binding(get_device_schema())
    proxy2 = DeviceProxy(binding=binding2, server_id="Test",
                         status=ProxyStatus.ONLINE)
    apply_default_configuration(proxy2.binding)

    dialog = CompareDeviceConfigurationsDialog(proxy1, proxy2)
    yield dialog


def test_selector_dialog(gui_app):
    topology = SystemTopology()
    h = deepcopy(system_hash())
    h["device.davyy"] = None
    h["device.davyy", ...] = {
        "host": "BIG_IRON",
        "type": "device",
        "capabilities": 0,
        "serverId": "otherserver",
        "classId": "FooClassCopy",
        "status": "ok",
        "interfaces": 0,
    }
    topology.initialize(h)
    with singletons(topology=topology):
        dialog = DeviceSelectorDialog("divvy")
        assert dialog.devices_combobox.count() == 1
        assert dialog.devices_combobox.currentText() == "davyy"


def test_compare_config_dialog(dialog):
    dialog._show_changes = False
    dialog._update_text()
    config1 = ("\nstate\nON"
               "\nstringProperty\nfoo"
               "\nreadOnlyProperty\n0"
               "\navailableScenes\n['scene']"
               "\nintProperty\n10"
               "\ninitOnlyString\nKarabo\n")
    assert dialog.reference_device_config.toPlainText() == config1

    config2 = ("\nstate\nON"
               "\nstringProperty\nfoo"
               "\nreadOnlyProperty\n0"
               "\navailableScenes\n['scene']\n")
    assert dialog.target_device_config.toPlainText() == config2

    # Show only reconfigurable properties.
    dialog.ui_show_only_reconfigurable.setChecked(True)
    dialog._update_text()

    reconfigurable1 = ("\nstate\nON"
                       "\nstringProperty\nfoo"
                       "\navailableScenes\n['scene']"
                       "\nintProperty\n10\n")
    assert dialog.reference_device_config.toPlainText() == reconfigurable1

    reconfigurable2 = ("\nstate\nON"
                       "\nstringProperty\nfoo"
                       "\navailableScenes\n['scene']\n")
    assert dialog.target_device_config.toPlainText() == reconfigurable2


def test_show_only_changes(dialog):
    assert dialog._show_changes
    config1 = ("\ninitOnlyString\nKarabo"
               "\nintProperty\n10\n")

    assert dialog.reference_device_config.toPlainText() == config1

    config2 = ("\ninitOnlyString""\nRemoved from configuration"
               "\nintProperty\nRemoved from configuration\n")

    assert dialog.target_device_config.toPlainText() == config2

    # Show only reconfigurable properties.
    dialog.ui_show_only_reconfigurable.setChecked(True)
    dialog._update_text()

    reconfigurable1 = ("\nintProperty\n10\n")
    assert dialog.reference_device_config.toPlainText() == reconfigurable1

    reconfigurable2 = ("\nintProperty\nRemoved from configuration\n")
    assert dialog.target_device_config.toPlainText() == reconfigurable2



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

from qtpy.QtCore import QSize
from qtpy.QtSvg import QSvgWidget
from qtpy.QtWidgets import QLabel, QWidget

from karabo.common.scenemodel.api import ImageRendererModel, create_base64image
from karabogui import icons
from karabogui.sceneview.widget.renderer import ImageRendererWidget
from karabogui.testing import GuiTestCase


class TestRenderers(GuiTestCase):

    def setUp(self):
        super().setUp()
        # The widget needs to have a shown parent in order to
        # test the sizes.
        self.parent = QWidget()
        self.parent.show()

    def tearDown(self):
        self.parent.destroy()
        self.parent = None

    def test_image_renderer(self):
        with self.subTest(msg="Test a valid svg file with renderer widget"):
            path = Path(icons.__file__).parent / "about.svg"
            with open(path, "rb") as fp:
                b = fp.read()

            image = create_base64image("svg", b)
            model = ImageRendererModel(
                x=0, y=0, image=image, width=100, height=100)
            widget = ImageRendererWidget(model=model, parent=self.parent)

            assert widget.size() == QSize(100, 100)
            assert isinstance(widget._internal_widget, QSvgWidget)
            # Valid renderer
            assert widget._internal_widget.renderer().isValid()
            widget.destroy()

        with self.subTest(msg="Test png file with renderer widget"):
            path = Path(icons.__file__).parent / "splash.png"
            with open(path, "rb") as fp:
                b = fp.read()

            image = create_base64image("png", b)
            model = ImageRendererModel(
                x=0, y=0, image=image, width=50, height=50)

            widget = ImageRendererWidget(model=model, parent=self.parent)
            assert isinstance(widget._internal_widget, QLabel)
            assert widget.size() == QSize(50, 50)
            widget.destroy()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 6, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from abc import abstractmethod

from qtpy.QtCore import QLine, QMargins, QPoint, QRect, QSize, Qt
from qtpy.QtGui import QBrush, QColor, QPainterPath, QPen, QPolygon, QTransform
from qtpy.QtWidgets import QDialog
from traits.api import (
    ABCHasStrictTraits, Bool, Constant, Float, Instance, Property, Tuple,
    cached_property)

from karabo.common.scenemodel.api import BaseShapeObjectData
from karabo.common.utils import get_arrowhead_points
from karabogui.dialogs.pen_dialogs import PenDialog
from karabogui.pathparser import Parser

from .const import (
    GRID_STEP, QT_PEN_CAP_STYLE_FROM_STR, QT_PEN_CAP_STYLE_TO_STR,
    QT_PEN_JOIN_STYLE_FROM_STR, QT_PEN_JOIN_STYLE_TO_STR, SCREEN_MAX_VALUE)

_BRUSH_ATTRS = ('fill', 'fill_opacity')
_PEN_ATTRS = ('stroke', 'stroke_opacity', 'stroke_linecap',
              'stroke_dashoffset', 'stroke_width', 'stroke_dasharray',
              'stroke_style', 'stroke_linejoin', 'stroke_miterlimit')


class BaseShape(ABCHasStrictTraits):
    """A shape base class
    """
    # The scene data model backing this shape
    model = Instance(BaseShapeObjectData)
    # The brush used by this shape
    brush = Property(Instance(QBrush),
                     depends_on=['model.' + attr for attr in _BRUSH_ATTRS])
    # The pen used by this shape
    pen = Property(Instance(QPen),
                   depends_on=['model.' + attr for attr in _PEN_ATTRS])
    # Is the shape currently shown?
    _hidden = Bool(False)

    @cached_property
    def _get_brush(self):
        brush = QBrush()
        if self.model.fill != "none":
            color = QColor(self.model.fill)
            color.setAlphaF(self.model.fill_opacity)
            brush.setStyle(Qt.SolidPattern)
            brush.setColor(color)
        return brush

    @cached_property
    def _get_pen(self):
        pen = QPen()
        model = self.model
        if model.stroke == "none" or model.stroke_width == 0:
            pen.setStyle(Qt.NoPen)
        else:
            c = QColor(model.stroke)
            c.setAlphaF(model.stroke_opacity)
            pen.setColor(c)
            pen.setCapStyle(QT_PEN_CAP_STYLE_FROM_STR[model.stroke_linecap])
            pen.setWidthF(model.stroke_width)
            pen.setDashOffset(model.stroke_dashoffset)

            if model.stroke_dasharray:
                pen.setDashPattern(model.stroke_dasharray)

            pen.setStyle(model.stroke_style)
            join_style = QT_PEN_JOIN_STYLE_FROM_STR[model.stroke_linejoin]
            pen.setJoinStyle(join_style)
            pen.setMiterLimit(model.stroke_miterlimit)
        return pen

    def edit(self, scene_view):
        """ Edits the pen and brush of the shape."""
        dialog = PenDialog(self.pen, self.brush, parent=scene_view)
        if dialog.exec() == QDialog.Rejected:
            return

        brush = dialog.brush
        if brush is not None and brush.style() == Qt.SolidPattern:
            fill = brush.color().name()
            fill_opacity = brush.color().alphaF()
        else:
            fill = 'none'
            fill_opacity = 1.0

        pen = dialog.pen
        self.model.trait_set(
            stroke='none' if pen.style() == Qt.NoPen else pen.color().name(),
            stroke_opacity=pen.color().alphaF(),
            stroke_linecap=QT_PEN_CAP_STYLE_TO_STR[pen.capStyle()],
            stroke_dashoffset=pen.dashOffset(),
            stroke_width=0 if pen.style() == Qt.NoPen else pen.widthF(),
            stroke_dasharray=pen.dashPattern(),
            stroke_style=pen.style(),
            stroke_linejoin=QT_PEN_JOIN_STYLE_TO_STR[pen.joinStyle()],
            stroke_miterlimit=pen.miterLimit(),
            fill=fill,
            fill_opacity=fill_opacity
        )

    def hide(self):
        self._hidden = True

    def show(self):
        self._hidden = False

    def is_visible(self):
        return not self._hidden

    @abstractmethod
    def draw(self, painter):
        """Needs to be reimplemented in inherited classes to draw the shape.
        """

    @abstractmethod
    def geometry(self):
        """Needs to be reimplemented in inherited classes to get the geometry
        for the bounding rectangle.
        """

    @abstractmethod
    def set_geometry(self, rect):
        """Needs to be reimplemented in inherited classes to set the geometry
        for the bounding rectangle.
        """

    @abstractmethod
    def translate(self, offset):
        """Needs to be reimplemented in inherited classes to move the shape.
        """

    def minimumSize(self):
        return QSize(0, 0)

    def maximumSize(self):
        return QSize(SCREEN_MAX_VALUE, SCREEN_MAX_VALUE)

    def sizeHint(self):
        return self.geometry()


class LineShape(BaseShape):
    """A line which can appear in a scene
    """
    # The line that we're drawing
    shape = Property(Instance(QLine), depends_on=['model.x1', 'model.x2',
                                                  'model.y1', 'model.y2'])

    # The geometry rect of the line. This includes margins that is independent
    # to the model values
    rect = Property(Instance(QRect), depends_on="shape")

    # There is a need to put margins to avoid mouse interaction problems.
    # This does not affect the model, but will most probably have an effect on
    # line shapes on a group/layout. This is an unavoidable evil.
    _margins = (GRID_STEP, GRID_STEP, GRID_STEP, GRID_STEP)
    _qmargins = Constant(QMargins(*_margins))

    # Record the vector direction on class instantiation. This is used to get
    # the correct line coords as we use its normalized rect in the scene.
    has_negative_width = Bool
    has_negative_height = Bool

    def _get_brush(self):
        """Reimplement the base-class property getter. No brushes here! """
        return None

    @cached_property
    def _get_shape(self):
        model = self.model
        return QLine(model.x1, model.y1, model.x2, model.y2)

    @cached_property
    def _get_rect(self):
        line = self.shape
        return QRect(line.p1(), line.p2())

    def _has_negative_width_default(self):
        model = self.model
        return model.x2 < model.x1

    def _has_negative_height_default(self):
        model = self.model
        return model.y2 < model.y1

    def draw(self, painter):
        """The line gets drawn.
        """
        painter.setPen(self.pen)
        painter.drawLine(self.shape)

    def geometry(self):
        # We have to return the normalized rect as this is used for mapping
        # the mouse position.
        return self.rect.normalized() + self._qmargins

    def set_geometry(self, rect):
        """Sets the effective geometry in the model which is without
           the margins and the normalization."""

        # Correct the rect first by reverting our changes:
        # 1. margins - we remove them
        rect -= self._qmargins

        # 2. normalization - we do it manually
        x1, y1, x2, y2 = rect.getCoords()
        if self.has_negative_width:
            x1, x2 = x2, x1
        if self.has_negative_height:
            y1, y2 = y2, y1

        self.model.trait_set(x1=x1, y1=y1, x2=x2, y2=y2)

    def translate(self, offset):
        x1, y1 = self.model.x1, self.model.y1
        x2, y2 = self.model.x2, self.model.y2
        xoff, yoff = offset.x(), offset.y()
        self.model.trait_set(x1=x1 + xoff, y1=y1 + yoff,
                             x2=x2 + xoff, y2=y2 + yoff)

    def minimumSize(self):
        """We use the margins for the minimum size. This just means that the
           effective rect has zero width/height"""
        left, top, right, bottom = self._margins
        return QSize(left + right, top + bottom)


class ArrowShape(LineShape):
    """Arrow has an additional arrowhead painted at the end of the line"""

    def draw(self, painter):
        super().draw(painter)
        model = self.model
        hp1 = QPoint(model.hx1, model.hy1)
        hp3 = QPoint(model.hx2, model.hy2)
        hp2 = QPoint(model.x2, model.y2)

        # Paint arrowhead as a polygon of three points
        painter.setBrush(QBrush(painter.pen().color()))
        painter.drawPolygon(QPolygon([hp1, hp2, hp3]))

    def translate(self, offset):
        super().translate(offset)
        hx1, hy1, hx2, hy2 = get_arrowhead_points(
            self.model.x1, self.model.y1, self.model.x2, self.model.y2)
        self.model.trait_set(hx1=hx1, hy1=hy1, hx2=hx2, hy2=hy2)

    def set_geometry(self, rect):
        super().set_geometry(rect)
        hx1, hy1, hx2, hy2 = get_arrowhead_points(
            self.model.x1, self.model.y1, self.model.x2, self.model.y2)
        self.model.trait_set(hx1=hx1, hy1=hy1, hx2=hx2, hy2=hy2)


class RectangleShape(BaseShape):
    """A rectangle which can appear in a scene
    """
    # The rectangle that we're drawing
    shape = Property(Instance(QRect),
                     depends_on=['model.x', 'model.y', 'model.height',
                                 'model.width'])

    @cached_property
    def _get_shape(self):
        model = self.model
        return QRect(model.x, model.y, model.width, model.height)

    def draw(self, painter):
        painter.setPen(self.pen)
        painter.setBrush(self.brush)
        painter.drawRect(self.shape)

    def geometry(self):
        return self.shape

    def set_geometry(self, rect):
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())

    def translate(self, offset):
        x, y = self.model.x, self.model.y
        self.model.trait_set(x=x + offset.x(), y=y + offset.y())


class PathShape(BaseShape):
    """A path which can appear in a scene
    """
    # The path object
    path = Instance(QPainterPath)
    shape = Property(Instance(QPainterPath), depends_on=["path", "transform"])
    _angle = Float(0.0)
    _offset = Tuple(0.0, 0.0)
    _scale = Float(1.0)
    transform = Property(Instance(QTransform),
                         depends_on=["_angle", "_offset", "_scale"])

    def _path_default(self):
        parser = Parser(self.model.svg_data)
        return parser.parse()

    @cached_property
    def _get_shape(self):
        return self.transform.map(self.path)

    def draw(self, painter):
        painter.setPen(self.pen)
        painter.setBrush(self.brush)
        painter.drawPath(self.shape)

    def geometry(self):
        return self.shape.boundingRect().toRect()

    def set_geometry(self, rect):
        pass

    def translate(self, offset):
        self._offset = offset.x(), offset.y()

    def rotate(self, angle):
        self._angle = -angle

    def scale(self, scale):
        self._scale = scale

    @cached_property
    def _get_transform(self):
        transform = QTransform()
        transform.translate(*self._offset)
        transform.scale(self._scale, self._scale)
        transform.rotate(self._angle)
        return transform



#############################################################################
# Author: <dennis.goeeries@xfel.eu>
# Created on February 12, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QPoint, Qt, Slot
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import (
    QAbstractItemView, QAction, QDialog, QHeaderView, QMenu, QTreeView)

from karabo.common.api import Capabilities
from karabogui import icons
from karabogui.access import AccessRole, access_role_allowed
from karabogui.const import TOOLTIP_INIT_CONFIG, TOOLTIP_RUNTIME_CONFIG
from karabogui.dialogs.api import (
    ConfigurationFromPastDialog, DeviceCapabilityDialog,
    InitConfigurationDialog)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.itemtypes import NavigationItemTypes
from karabogui.request import get_scene_from_server
from karabogui.singletons.api import get_manager, get_selection_tracker
from karabogui.widgets.popup import PopupWidget

from .device_filter_model import DeviceFilterModel
from .device_model import DeviceTreeModel
from .tools import DeviceSceneHandler


class DeviceTreeView(QTreeView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("deviceTreeView")
        self._selected_proxy = None  # A BaseDeviceProxy

        model = DeviceTreeModel(parent=self)
        proxy_model = DeviceFilterModel(parent=self,
                                        source_model=model)
        self.setModel(proxy_model)
        self.setSelectionModel(proxy_model.selectionModel)
        proxy_model.signalItemChanged.connect(self.onSelectionChanged)
        proxy_model.modelReset.connect(self.resetExpand)
        header = self.header()
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        # Prevent drag reorder of the header
        header.setSectionsMovable(False)

        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.handler_list = [DeviceSceneHandler()]
        self.expanded = False
        self.popupWidget = None
        self.header().sectionDoubleClicked.connect(self.onDoubleClickHeader)
        self.setDragEnabled(True)

        # Setup the context menu
        self._setup_context_menu()
        self.customContextMenuRequested.connect(
            self.onCustomContextMenuRequested)
        self.setUniformRowHeights(True)

    def _setup_context_menu(self):
        """Setup the context menu for the device topology"""
        self.setContextMenuPolicy(Qt.CustomContextMenu)

        self.menu = QMenu(self)
        text = "About"
        self.ac_about = QAction(icons.about, "About", self)
        self.ac_about.setStatusTip(text)
        self.ac_about.setToolTip(text)
        self.ac_about.triggered.connect(self.onAbout)

        text = "Runtime configuration"
        self.ac_config_past = QAction(icons.clock, text, self)
        self.ac_config_past.setStatusTip(text)
        self.ac_config_past.setToolTip(TOOLTIP_RUNTIME_CONFIG)
        self.ac_config_past.triggered.connect(self.onGetConfigurationFromPast)

        text = "Init configuration"
        self.ac_config_name = QAction(icons.initconfig, text, self)
        self.ac_config_name.setStatusTip(text)
        self.ac_config_name.setToolTip(TOOLTIP_INIT_CONFIG)
        self.ac_config_name.triggered.connect(self.onGetInitConfiguration)

        text = "Open device scene..."
        self.ac_open_device_scene = QAction(text, self)
        self.ac_open_device_scene.setStatusTip(text)
        self.ac_open_device_scene.setToolTip(text)
        self.ac_open_device_scene.setVisible(False)
        self.ac_open_device_scene.triggered.connect(self.onOpenDeviceScene)

        text = "Shutdown device"
        self.ac_kill_device = QAction(icons.delete, text, self)
        self.ac_kill_device.setStatusTip(text)
        self.ac_kill_device.setToolTip(text)
        self.ac_kill_device.triggered.connect(self.onKillInstance)

        self.menu.addAction(self.ac_config_past)
        self.menu.addAction(self.ac_config_name)
        self.menu.addSeparator()
        self.menu.addAction(self.ac_kill_device)
        self.menu.addAction(self.ac_open_device_scene)
        self.menu.addSeparator()
        self.menu.addAction(self.ac_about)

    def indexInfo(self, index=None):
        """Return the info about the index.

        Defaults to the current index if index is None.
        """
        if index is None:
            index = self.currentIndex()
        return self.model().indexInfo(index)

    def currentIndex(self):
        return self.model().currentIndex()

    def clear(self):
        self._selected_proxy = None
        self.clearSelection()
        self.model().clear()

    def scrollTo(self, index, hint=QAbstractItemView.EnsureVisible):
        """Reimplementation of the Qt function
        """
        # XXX: We used to expand the index here!
        super().scrollTo(index, hint)

    # ----------------------------
    # Events

    def mouseDoubleClickEvent(self, event):
        index = self.currentIndex()
        node = self.model().index_ref(index)
        # QModelIndexes tend to outlive the node objects which they reference!
        if node is None:
            return
        info = node.info()
        for handler in self.handler_list:
            if handler.can_handle(info):
                handler.handle(info)
                event.accept()
                return

        super().mouseDoubleClickEvent(event)

    # ----------------------------
    # Slots

    @Slot(str, object)
    def onSelectionChanged(self, item_type, proxy):
        """Called by the data model when an item is selected
        """
        self._selected_proxy = proxy

        # Grab control of the global selection
        get_selection_tracker().grab_selection(self.model().selectionModel)

        # Tell the configurator
        if item_type not in ('class', 'device'):
            # servers and hosts clear the configurator
            proxy = None
        broadcast_event(KaraboEvent.ShowConfiguration, {'proxy': proxy})

    @Slot()
    def onAbout(self):
        index = self.currentIndex()
        node = self.model().index_ref(index)
        if node is None:
            return
        if self.popupWidget is None:
            self.popupWidget = PopupWidget(parent=self)
        self.popupWidget.setInfo(node.attributes)

        pos = QCursor.pos()
        pos.setX(pos.x() + 10)
        pos.setY(pos.y() + 10)
        self.popupWidget.move(pos)
        self.popupWidget.show()

    @Slot()
    def onOpenDeviceScene(self):
        info = self.indexInfo()
        dialog = DeviceCapabilityDialog(
            device_id=info.get('deviceId', ''),
            capability=Capabilities.PROVIDES_SCENES,
            parent=self)
        if dialog.exec() == QDialog.Accepted:
            device_id = dialog.device_id
            scene_name = dialog.capa_name
            get_scene_from_server(device_id, scene_name=scene_name)

    @Slot()
    def onGetConfigurationFromPast(self):
        info = self.indexInfo()
        device_id = info['deviceId']
        dialog = ConfigurationFromPastDialog(instance_id=device_id,
                                             parent=self)
        dialog.move(QCursor.pos())
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    @Slot()
    def onGetInitConfiguration(self):
        info = self.indexInfo()
        device_id = info.get('deviceId')
        dialog = InitConfigurationDialog(instance_id=device_id,
                                         parent=self)
        dialog.move(QCursor.pos())
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    @Slot(QPoint)
    def onCustomContextMenuRequested(self, pos):
        def _test_mask(mask, bit):
            return (mask & bit) == bit

        info = self.indexInfo()
        node_type = info.get("type", NavigationItemTypes.UNDEFINED)
        if node_type is NavigationItemTypes.DEVICE:
            # Killing services is access level dependent!
            enable = access_role_allowed(AccessRole.SERVICE_EDIT)
            self.ac_kill_device.setEnabled(enable)
            has_scenes = _test_mask(info.get("capabilities", 0),
                                    Capabilities.PROVIDES_SCENES)
            self.ac_open_device_scene.setVisible(has_scenes)
            self.menu.exec(QCursor.pos())

    @Slot()
    def onKillInstance(self):
        info = self.indexInfo()
        node_type = info.get('type')
        manager = get_manager()

        if node_type is NavigationItemTypes.DEVICE:
            deviceId = info.get('deviceId')
            manager.shutdownDevice(deviceId, parent=self)

    @Slot()
    def onDoubleClickHeader(self):
        if self.expanded:
            self.collapseAll()
        else:
            self.expandAll()

        index = self.model().currentIndex()
        if index.isValid():
            self.scrollTo(index)

    @Slot()
    def resetExpand(self):
        self.expanded = False

    def collapseAll(self):
        self.expanded = False
        super().collapseAll()

    def expandAll(self):
        self.expanded = True
        super().expandAll()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import Qt

from karabo.common.scenemodel.api import CheckBoxModel
from karabo.common.states import State
from karabo.native import Bool, Configurable, String
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..checkbox import EditableCheckBox


class Object(Configurable):
    state = String(defaultValue=State.ON)
    prop = Bool(allowedStates=[State.ON])


@pytest.fixture
def checkbox_setup(gui_app):
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    controller = EditableCheckBox(proxy=proxy, model=CheckBoxModel())
    controller.create(None)
    assert controller.widget is not None

    yield controller, proxy

    controller.destroy()
    assert controller.widget is None


def test_checkbox_state_update(checkbox_setup):
    controller, proxy = checkbox_setup
    set_proxy_value(proxy, "state", "CHANGING")
    assert controller.widget.isEnabled() is False
    set_proxy_value(proxy, "state", "ON")
    assert controller.widget.isEnabled() is True


def test_checkbox_set_value(checkbox_setup):
    controller, proxy = checkbox_setup
    set_proxy_value(proxy, "prop", True)
    assert controller.widget.checkState() == Qt.Checked


def test_checkbox_edit_value(checkbox_setup):
    controller, proxy = checkbox_setup
    controller.widget.setCheckState(Qt.Checked)
    assert proxy.edit_value

    controller.widget.setCheckState(Qt.Unchecked)
    assert not proxy.edit_value



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Callable, List

from karabogui import icons
from karabogui.graph.common.api import (
    AuxPlots, BaseToolsetController, create_button)


def aux_plots_factory(tool):
    button = None
    if tool is AuxPlots.ProfilePlot:
        button = create_button(icon=icons.beamProfile,
                               checkable=True,
                               tooltip="Beam Profile")
    elif tool == AuxPlots.Histogram:
        button = create_button(icon=icons.histogram,
                               checkable=True,
                               tooltip="Histogram")
    return button


class AuxPlotsToolset(BaseToolsetController):
    tools = List([AuxPlots.ProfilePlot, AuxPlots.Histogram])
    factory = Callable(aux_plots_factory)

    default_tool = AuxPlots.NoPlot

    def select(self, tool):
        """The toolset has can have one or more buttons, with check states
           being exclusive. When a button is unchecked, the toolset returns
           ROITool.NoROI. When selected, on the other hand, the other button
           (if existing), should be unchecked."""

        # Toolset has two buttons, check state are then exclusive
        prof_button = self.buttons[AuxPlots.ProfilePlot]
        hist_button = self.buttons[AuxPlots.Histogram]

        # Uncheck the other button
        if not prof_button.isChecked() and not hist_button.isChecked():
            tool = AuxPlots.NoPlot
        elif tool == AuxPlots.ProfilePlot and prof_button.isChecked():
            hist_button.setChecked(False)
        elif tool == AuxPlots.Histogram and hist_button.isChecked():
            prof_button.setChecked(False)

        super().select(tool)

    def check(self, tool):
        if tool == AuxPlots.NoPlot:
            self.uncheck_all()
            return

        super().check(tool)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on May 7, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QEvent, QMargins
from qtpy.QtWidgets import (
    QApplication, QStyle, QStyledItemDelegate, QStyleOptionButton)


class ButtonState:
    PRESSED = QStyle.State_Enabled | QStyle.State_Sunken
    ENABLED = QStyle.State_Enabled | QStyle.State_Raised | QStyle.State_Off
    DISABLED = QStyle.State_On


MARGINS = QMargins(4, 4, 4, 4)


class TableButtonDelegate(QStyledItemDelegate):
    """A QStyledItemDelegate for slot buttons
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._enabled = True
        self._button_states = {}

    def paint(self, painter, option, index):
        """Reimplemented function of QStyledItemDelegate."""
        self._draw_button(painter, option, index)

    def editorEvent(self, event, model, option, index):
        """Reimplemented function of QStyledItemDelegate."""
        handled_types = (QEvent.MouseButtonPress, QEvent.MouseButtonRelease,
                         QEvent.MouseMove)
        if event.type() in handled_types:
            self._handle_event_state(event, option, index)
            return True
        return super().editorEvent(
            event, model, option, index)

    def _draw_button(self, painter, option, index):
        if not self.isEnabled(index):
            state = ButtonState.DISABLED
        else:
            key = (index.row(), index.column())
            state = self._button_states.get(key, ButtonState.ENABLED)
        button = QStyleOptionButton()
        button.state = state
        button.rect = option.rect
        button.text = self.get_button_text(index) or "No text"
        button.features = QStyleOptionButton.None_
        QApplication.style().drawControl(QStyle.CE_PushButton, button, painter)

    def _handle_event_state(self, event, option, index):
        if not self.isEnabled(index):
            # We are disabled but still clickable. We prevent that our action
            # is executed here!
            return
        key = (index.row(), index.column())
        state = self._button_states.get(key, ButtonState.ENABLED)
        # Remove some margins in case the mouse moves to bail out the pressed
        # button state!
        rect = option.rect
        rect = rect.marginsRemoved(MARGINS)
        if rect.contains(event.pos()):
            if event.type() == QEvent.MouseButtonPress:
                state = ButtonState.PRESSED
        else:
            state = ButtonState.ENABLED
        if (state == ButtonState.PRESSED and
                event.type() == QEvent.MouseButtonRelease):
            state = ButtonState.ENABLED
            self.click_action(index)
        self._button_states[key] = state

    # -----------------------------------------------------------------------
    # Public interface

    def isEnabled(self, index=None):
        """Return if the ButtonDelegate is enabled for `index`

        This method can be used to subclass for a specific index
        """
        return self._enabled

    def setEnabled(self, enable):
        """Set globally if the ButtonDelegate is enabled"""
        self._enabled = enable
        self._button_states = {}

    # -----------------------------------------------------------------------
    # Abstract interface

    def get_button_text(self, index):
        """Subclass this method to define the button text for `index`"""

    def click_action(self, index):
        """Subclass this method to define the action for `index`"""



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on April 23, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op

from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox
from traits.api import Undefined

from karabo.native import (
    AccessMode, Assignment, Hash, create_html_hash, writeXML)
from karabogui import icons, messagebox
from karabogui.binding.api import (
    ProjectDeviceProxy, extract_configuration, get_config_changes,
    iterate_binding)
from karabogui.singletons.api import get_config
from karabogui.util import SignalBlocker, getSaveFileName

from .utils import get_dialog_ui


class ConfigurationFromPastPreview(QDialog):
    """A simple configuration preview for configuration from name and time

    :param title: The title of the dialog
    :param info: The text info for the field (sub title). If not info is
                 provided, the field is invisible
    :param proxy: The proxy instance, can be from project or topology
    :param parent: The dialog parent
    """

    def __init__(self, title, info, configuration, proxy, parent=None):
        super().__init__(parent=parent)
        self._currently_reconfigurable = ""
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        ui_file = get_dialog_ui("configuration_past_preview.ui")
        uic.loadUi(ui_file, self)
        self.setWindowTitle(title)
        flags = Qt.WindowCloseButtonHint | Qt.WindowStaysOnTopHint
        self.setWindowFlags(self.windowFlags() | flags)

        if info is None:
            self.ui_info.setVisible(False)
        else:
            status_text = "online" if proxy.online else "offline"
            info = f"{info} The device is {status_text}."
            self.ui_info.setText(info)

        self._show_changes = False
        self.ui_swap.setIcon(icons.change)
        self.ui_swap.clicked.connect(self._swap_view)
        ok_button = self.ui_buttonBox.button(QDialogButtonBox.Ok)
        ok_button.clicked.connect(self.accept)

        cancel_button = self.ui_buttonBox.button(QDialogButtonBox.Cancel)
        cancel_button.clicked.connect(self.reject)

        self.ui_save_configuration.clicked.connect(
            self.save_configuration_to_file)
        self.ui_save_configuration.setIcon(icons.save)

        self.proxy = proxy
        self.configuration = configuration
        self._configuration_by_access_mode()
        self._show_configuration()
        self._load_configuration_changes()
        self.ui_synchronize_bars.toggled.connect(
            self._synchronize_toggled)
        self._synchronize_toggled(True)
        self.ui_hide_readonly.toggled.connect(self._show_configuration_changes)
        self._show_configuration_changes(hide_readonly=False)
        self.ui_reconfigurable_only.setVisible(self.proxy.online)
        self.ui_config_count.setVisible(self.proxy.online)
        self.ui_reconfigurable_only.toggled.connect(
            self._show_reconfigurable_only)

    def _configuration_by_access_mode(self):
        reconfig_text = "No Reconfigurable Property Available"
        if len(self.proxy.binding.value):
            if self.proxy.online:
                reconfig, current_reconfig = \
                    self._extract_online_reconfigurable()
                self._currently_reconfigurable = create_html_hash(
                    current_reconfig)
                allowed_count = len(current_reconfig.paths(intermediate=False))
                reconfig_count = len(reconfig.paths(intermediate=False))
                text = (f"Showing {reconfig_count} reconfigurable parameters "
                        f"and {allowed_count} are currently allowed.")
                self.ui_config_count.setText(text)
            else:
                reconfig = self._extract_offline_reconfigurable()
            reconfig_text = create_html_hash(reconfig)
            self._reconfigurable_props = reconfig_text
        self.ui_text_info_configurable.setHtml(reconfig_text)

    def _load_configuration_changes(self):
        """Store the changed configuration """
        if not len(self.proxy.binding.value):
            # Note: We are protected by the menu bar that does not allow us
            # to be launched in this case. However, we protect again here...
            self.all_changes_existing = self.configurable_changes_existing = (
                "No schema available to extract a configuration")
            self.configurable_changes_existing = self.all_changes_retrieved = (
                "No schema available for comparing configurations!")
            return

        project = isinstance(self.proxy, ProjectDeviceProxy)
        existing = extract_configuration(self.proxy.binding)
        changes_a, changes_b = get_config_changes(existing, self.configuration,
                                                  project)
        self.all_changes_existing = create_html_hash(
            changes_a, include_attributes=False)
        self.all_changes_retrieved = create_html_hash(
            changes_b, include_attributes=False)

        configurable_changes_a, configurable_changes_b = \
            self._skip_read_only_changes(changes_a, changes_b)
        self.configurable_changes_existing = create_html_hash(
            configurable_changes_a, include_attributes=False)
        self.configurable_changes_retrieved = create_html_hash(
            configurable_changes_b, include_attributes=False)

    def _show_configuration(self):
        html = create_html_hash(self.configuration, include_attributes=False)
        self.ui_text_info_all.setHtml(html)

    def _skip_read_only_changes(self, changes_a: Hash, changes_b: Hash) -> \
            tuple[Hash, Hash]:
        """Remove the read only properties from the given Configurations"""
        read_only_changes_a, read_only_changes_b = Hash(), Hash()
        paths_a = changes_a.paths(intermediate=False)
        paths_b = changes_b.paths(intermediate=False)
        keys = sorted(set(paths_a).union(set(paths_b)))
        for key in keys:
            if key in self._read_only_props:
                continue
            read_only_changes_a[key] = changes_a[key]
            read_only_changes_b[key] = changes_b[key]
        return read_only_changes_a, read_only_changes_b

    def _extract_offline_reconfigurable(self):
        """
        Extract the readonly and reconfigurable property from the configuration
        for an offline device. Reconfigurable properties includes the init only
        properties as well.
        """
        binding = self.proxy.binding
        read_only = Hash()
        reconfigurable = Hash()

        for key, node in iterate_binding(binding):
            value = self.configuration.get(key, None)
            if value is None or value is Undefined:
                continue

            accessMode = node.accessMode
            if accessMode in (
                    AccessMode.RECONFIGURABLE, AccessMode.INITONLY) and (
                    node.assignment != Assignment.INTERNAL):
                reconfigurable[key] = value
            else:
                read_only[key] = value
        self._read_only_props = read_only
        return reconfigurable

    def _extract_online_reconfigurable(self):
        """
        Extract the readonly and reconfigurable property from the configuration
        for an online device. Consider the 'allowedStates' attribute to
        determine if the property is reconfigurable or not allowed in the
        current state of the device.
        """

        device_state = self.proxy.state_binding
        binding = self.proxy.binding
        read_only = Hash()
        reconfigurable = Hash()
        currently_reconfigurable = Hash()

        for key, node in iterate_binding(binding):
            value = self.configuration.get(key, None)
            if value is None or value is Undefined:
                continue

            accessMode = node.accessMode
            if accessMode == AccessMode.RECONFIGURABLE:
                reconfigurable[key] = value
                if node.is_allowed(state=device_state):
                    currently_reconfigurable[key] = value
            else:
                read_only[key] = value
        self._read_only_props = read_only
        return reconfigurable, currently_reconfigurable

    # ---------------------------------------------------------------------
    # Slot Interface

    @Slot(bool)
    def _synchronize_toggled(self, enabled):
        """Internal method to synchronize the scrollbars"""
        existing = self.ui_existing.verticalScrollBar()
        retrieved = self.ui_retrieved.verticalScrollBar()
        if enabled:
            existing.valueChanged.connect(self.existing_bar_moved)
            retrieved.valueChanged.connect(self.retrieved_bar_moved)
        else:
            existing.valueChanged.disconnect(self.existing_bar_moved)
            retrieved.valueChanged.disconnect(self.retrieved_bar_moved)

    @Slot(int)
    def existing_bar_moved(self, value):
        with SignalBlocker(self.ui_retrieved):
            self.ui_retrieved.verticalScrollBar().setValue(value)

    @Slot(int)
    def retrieved_bar_moved(self, value):
        with SignalBlocker(self.ui_existing):
            self.ui_existing.verticalScrollBar().setValue(value)

    @Slot()
    def _swap_view(self):
        self._show_changes = not self._show_changes
        text = "Show Configuration" if self._show_changes else "Show Changes"
        self.ui_swap.setText(text)
        self.ui_stack_widget.setCurrentIndex(int(self._show_changes))

        text = "Changes" if self._show_changes else "Retrieved Configuration"
        self.ui_show.setText(text)

    @Slot()
    def save_configuration_to_file(self):
        if self.configuration.empty():
            messagebox.show_error("The configuration is empty and cannot "
                                  "be saved", title='No configuration',
                                  parent=self)
            return

        config = get_config()
        path = config['data_dir']
        directory = path if path and op.isdir(path) else ""

        class_id = self.configuration.get("classId", "unknown-class")

        # Use deviceId for the name!
        default_name = self.configuration.get('deviceId')
        if default_name:
            default_name = default_name.replace('/', '-') + '.xml'
        else:
            default_name = class_id + '.xml'

        filename = getSaveFileName(caption="Save configuration as",
                                   filter="Configuration (*.xml)",
                                   suffix="xml",
                                   directory=directory,
                                   selectFile=default_name,
                                   parent=self)
        if not filename:
            return

        config = Hash(class_id, self.configuration)

        # Save configuration to file
        with open(filename, 'w') as fp:
            writeXML(config, fp)

        # save the last config directory
        get_config()['data_dir'] = op.dirname(filename)

    @Slot(bool)
    def _show_configuration_changes(self, hide_readonly: bool) -> None:
        """Show the configuration in the widgets.Optionally hides the
        readonly properties"""
        if hide_readonly:
            html_a = self.configurable_changes_existing
            html_b = self.configurable_changes_retrieved
        else:
            html_a = self.all_changes_existing
            html_b = self.all_changes_retrieved
        self.ui_existing.setHtml(html_a)
        self.ui_retrieved.setHtml(html_b)

    @Slot(bool)
    def _show_reconfigurable_only(self, toggled: bool) -> None:
        """
        For online device, show either only the properties that are
        reconfigurable in the current state of the device or all
        reconfigurable properties.
        """
        text = (self._currently_reconfigurable if toggled else
                self._reconfigurable_props)
        self.ui_text_info_configurable.setHtml(text)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import Qt

from karabo.common.scenemodel.api import DisplayProgressBarModel
from karabo.native import Configurable, Float, Int8
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..progressbar import NULL_RANGE, PROGRESS_MAX, DisplayProgressBar


class Object(Configurable):
    prop = Float(minInc=-2.0, maxInc=4.0)


class ObjectWithoutLimits(Configurable):
    prop = Int8()


@pytest.fixture
def progressbar_setup(gui_app):
    schema = Object.getClassSchema()
    model = DisplayProgressBarModel(is_vertical=True)
    proxy = get_class_property_proxy(schema, "prop")
    controller = DisplayProgressBar(proxy=proxy, model=model)
    controller.create(None)
    yield controller, proxy, model

    # teardown
    controller.destroy()
    assert controller.widget is None


def test_widget(progressbar_setup):
    controller, _, _ = progressbar_setup
    assert controller.widget.minimum() == 0
    assert controller.widget.maximum() == PROGRESS_MAX


def test_orientation(progressbar_setup):
    controller, _, _ = progressbar_setup
    assert controller.widget.orientation() == Qt.Vertical

    # Trigger the orientation changing action
    action = controller.widget.actions()[0]
    assert action.text() == "Change Orientation"

    action.trigger()
    assert controller.widget.orientation() == Qt.Horizontal


def test_set_value(progressbar_setup):
    controller, proxy, _ = progressbar_setup
    # value range is [-2, 4]; 1.0 is the middle
    set_proxy_value(proxy, "prop", 1.0)
    assert controller.widget.value() == PROGRESS_MAX * 0.5


def test_no_limits_messagebox(gui_app):
    schema = ObjectWithoutLimits.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    model = DisplayProgressBarModel(is_vertical=True)
    controller = DisplayProgressBar(proxy=proxy, model=model)
    controller.create(None)

    assert controller.widget.minimum() == 0
    assert controller.widget.maximum() == 0
    assert controller._value_factors == NULL_RANGE

    # tear down
    controller.destroy()
    assert controller.widget is None



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on June 3, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox

import karabogui.icons as icons
from karabo.native import Hash, create_html_hash, has_changes

from .utils import get_dialog_ui


def config_changes(a, b):
    """Compare two `Hash` configurations `a` and `b`"""
    changes = Hash()
    for key, a_value, _ in Hash.flat_iterall(a):
        if key not in b:
            changes[key] = "Value was removed ..."
            continue
        b_value = b[key]
        if has_changes(a_value, b_value):
            changes[key] = b_value
    for key, b_value, _ in Hash.flat_iterall(b):
        if key not in a:
            changes[key] = b_value

    return changes


class ConfigComparisonDialog(QDialog):
    """A simple configuration comparison dialog

    :param title: The title of the dialog
    :param old: The existing `old` configuration for comparison
    :param new: The `new` configuration for comparison
    :param parent: The dialog parent
    """

    def __init__(self, title, old, new, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        uic.loadUi(get_dialog_ui("configuration_comparison.ui"), self)
        self.setWindowTitle(title)
        flags = Qt.WindowCloseButtonHint | Qt.WindowStaysOnTopHint
        self.setWindowFlags(self.windowFlags() | flags)

        text = ("Note: The configuration ONLY reflects property "
                "values that are different from the device default.")
        self.ui_info.setText(text)

        ok_button = self.ui_buttonBox.button(QDialogButtonBox.Ok)
        ok_button.clicked.connect(self.accept)

        cancel_button = self.ui_buttonBox.button(QDialogButtonBox.Cancel)
        cancel_button.clicked.connect(self.reject)

        self._show_comparison = False
        self.ui_swap.setIcon(icons.change)
        self.ui_swap.clicked.connect(self._swap_view)
        ok_button = self.ui_buttonBox.button(QDialogButtonBox.Ok)
        ok_button.clicked.connect(self.accept)

        cancel_button = self.ui_buttonBox.button(QDialogButtonBox.Cancel)
        cancel_button.clicked.connect(self.reject)

        html_a = create_html_hash(old)
        html_b = create_html_hash(new)
        self.ui_config_existing.setHtml(html_a)
        self.ui_config_new.setHtml(html_b)

        changes = config_changes(old, new)
        self.ui_changes.setHtml(create_html_hash(changes))

    # ---------------------------------------------------------------------
    # Slot Interface

    @Slot()
    def _swap_view(self):
        self._show_comparison = not self._show_comparison
        text = ("Show Configuration" if self._show_comparison
                else "Show Comparison")
        self.ui_swap.setText(text)
        self.ui_stack_widget.setCurrentIndex(int(self._show_comparison))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from copy import deepcopy

import numpy as np
from qtpy.QtGui import QBrush

from karabo.common.const import (
    KARABO_SCHEMA_ACCESS_MODE, KARABO_SCHEMA_DISPLAY_TYPE)
from karabo.native import AccessMode, Hash
from karabogui.binding.api import BoolBinding, StringBinding
from karabogui.controllers.table.utils import (
    create_brushes, get_button_attributes, has_confirmation,
    is_state_display_type, is_writable_binding, list2string, parse_table_link,
    quick_table_copy, string2list)


def test_displayType_state():
    string = StringBinding(attributes={KARABO_SCHEMA_DISPLAY_TYPE: "State"})
    assert is_state_display_type(string)
    string = StringBinding(attributes={KARABO_SCHEMA_DISPLAY_TYPE: "Alarm"})
    assert not is_state_display_type(string)


def test_is_writable_string():
    string = StringBinding(
        attributes={KARABO_SCHEMA_ACCESS_MODE: AccessMode.RECONFIGURABLE})
    assert is_writable_binding(string)
    # Test that is works also by specificing a string binding
    assert is_writable_binding(string, StringBinding)

    string = StringBinding(
        attributes={KARABO_SCHEMA_ACCESS_MODE: AccessMode.READONLY})
    assert not is_writable_binding(string)

    boolean = BoolBinding(
        attributes={KARABO_SCHEMA_ACCESS_MODE: AccessMode.RECONFIGURABLE})
    assert is_writable_binding(boolean)
    assert not is_writable_binding(boolean, StringBinding)


def test_convert_string_list():
    value = ""
    assert string2list(value) == []
    value = "2"
    assert string2list(value) == ["2"]
    value = "2,2,3"
    assert string2list(value) == ["2", "2", "3"]
    # Only for testing. Prevented by delegates later.
    value = "2,2,    3"
    assert string2list(value) == ["2", "2", "3"]
    value = ", , ,"
    assert string2list(value) == ["", "", "", ""]
    value = ",2,"
    assert string2list(value) == ["", "2", ""]


def test_convert_list_string():
    value = None
    assert list2string(value) == ""
    value = ["2", "4.2"]
    assert list2string(value) == "2,4.2"
    value = ["2", "2", "3"]
    assert list2string(value) == "2,2,3"


def test_create_brushes():
    display_string = "TableColor|offline=red&online=green&default=blue"
    default, brushes = create_brushes(display_string)
    assert default is not None
    assert isinstance(default, QBrush)
    assert "offline" in brushes
    assert "online" in brushes
    assert isinstance(brushes["online"], QBrush)
    assert isinstance(brushes["offline"], QBrush)

    # Wonky typos, e.g. ? as uri identifier
    display_string = "TableColor?offline=red&online=green&default=blue"
    default, brushes = create_brushes(display_string)
    assert default is None
    assert not brushes

    # Wrong delimiter, no exceptions
    display_string = "TableColor|offline=red%online=green%nodefault=blue"
    default, brushes = create_brushes(display_string)
    assert default is None
    assert len(brushes) == 1
    assert "offline" in brushes

    # No default brush
    display_string = "TableColor|offline=red&online=green&nodefault=blue"
    default, brushes = create_brushes(display_string)
    assert default is None
    assert "offline" in brushes
    assert "online" in brushes
    assert "nodefault" in brushes


def test_has_confirmation():
    confirm = has_confirmation("TableColor|confirmation=1")
    assert confirm is True
    # No caseless hickups, we have a contract
    confirm = has_confirmation("TableColor|Confirmation=1")
    assert confirm is False
    confirm = has_confirmation("TableColor|confirmation=1&icon=0")
    assert confirm is True
    confirm = has_confirmation("TableColor|confirmation=0")
    assert confirm is False
    confirm = has_confirmation("TableColor|confirmation")
    assert confirm is False


def test_get_button_attributes():
    attrs = get_button_attributes("TableColor|confirmation=1")
    assert attrs.get("confirmation") == "1"
    attrs = get_button_attributes("TableSelection|Confirmation=1&")
    assert attrs.get("Confirmation") == "1"
    attrs = get_button_attributes("TableBoolButton|confirmation=1&icon=0")
    assert attrs.get("confirmation") == "1"
    assert attrs.get("icon") == "0"


def test_table_scheme_parsing():
    action, options = parse_table_link(
        "deviceScene|device_id=CAM&name=scene")
    assert action == "deviceScene"
    assert options["device_id"] == "CAM"
    assert options["name"] == "scene"

    action, options = parse_table_link("url|https://www.xfel.eu")
    assert action == "url"
    assert options == "https://www.xfel.eu"


def test_copy_hash():
    h = Hash("boolean", False, "string", "ginger", "int", 1, "float", 2.1,
             "list", [1, 2], "array", np.array([1, 2, 3]), "emptylist", [],
             "emptyarray", np.array([]))

    SIMPLES = ["string", "float", "boolean"]

    for func in (quick_table_copy, deepcopy):
        r = func(h)
        assert r.fullyEqual(h)
        hv = h["list"]
        rv = r["list"]
        assert id(hv) != id(rv)
        # Same id, but immutable objects
        assert id(hv[0]) == id(rv[0])
        hv = h["array"]
        rv = r["array"]
        assert id(hv) != id(rv)
        hv = h["emptylist"]
        rv = r["emptylist"]
        assert id(hv) != id(rv)
        hv = h["emptyarray"]
        rv = r["emptyarray"]
        assert id(hv) != id(rv)

        for simple in SIMPLES:
            hv = h[simple]
            rv = r[simple]
            assert id(hv) == id(rv)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QPoint, QRect
from qtpy.QtWidgets import QDialog, QWidget

from karabo.common.scenemodel.api import LabelModel
from karabogui.sceneview.widget.label import LabelWidget
from karabogui.testing import GuiTestCase


class TestLabelWidget(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.model = LabelModel(
            text="Label Text", foreground="black",
            x=0, y=0, width=100, height=100, alignh=1)

        self.main_widget = QWidget()
        self.widget = LabelWidget(model=self.model,
                                  parent=self.main_widget)
        self.main_widget.show()

    def tearDown(self):
        super().tearDown()
        self.main_widget = None
        self.widget = None

    # -----------------------------------------------------------------------
    # Actual tests

    def test_basics(self):
        model_rect = QRect(self.model.x, self.model.y,
                           self.model.width, self.model.height)
        self.assertTrue(self.widget.rect() == model_rect)
        # Actions
        self.assertEqual(len(self.widget.actions()), 1)
        edit_action = self.widget.actions()[0]
        self.assertEqual(edit_action.text(), "Edit Label")

        text_before = "Label Text"
        self.assertEqual(self.widget.text(), text_before)
        self.assertEqual(self.model.text, text_before)

        path = "karabogui.sceneview.widget.label.TextDialog"
        with mock.patch(path) as dialog:
            dialog.exec.return_value = QDialog.Accepted
            text_after = "This is a label text"
            dialog().label_model = LabelModel(text=text_after)
            edit_action.trigger()
            self.assertEqual(self.model.text, text_after)

        # Make sure, basic interface is there
        self.widget.add_proxies(None)
        self.widget.apply_changes()
        self.widget.decline_changes()
        self.widget.destroy()
        self.widget.set_visible(True)
        self.widget.update_global_access_level(None)
        self.assertIsNotNone(self.widget)

        # Geometry
        rect = QRect(2, 10, 2, 10)
        self.assertFalse(self.widget.geometry() == rect)
        self.widget.set_geometry(rect)
        self.assertTrue(self.widget.geometry() == rect)

        # Translation
        self.assertEqual(self.widget.pos(), QPoint(2, 10))
        self.widget.translate(QPoint(10, 0))
        self.assertEqual(self.widget.pos(), QPoint(12, 10))


if __name__ == "__main__":
    main()



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 27, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op
import re
from functools import partial
from io import StringIO

from qtpy.QtWidgets import QAction, QDialog, QMenu, QMessageBox
from traits.api import Instance, String, on_trait_change

from karabo.common.api import walk_traits_object
from karabo.common.project.api import MacroModel, read_macro, write_macro
from karabogui import icons, messagebox
from karabogui.access import AccessRole, access_role_allowed
from karabogui.binding.api import ProxyStatus
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts,
    unregister_from_broadcasts)
from karabogui.indicators import get_project_device_status_icon
from karabogui.itemtypes import ProjectItemTypes
from karabogui.project.dialog.object_handle import (
    ObjectDuplicateDialog, ObjectEditDialog)
from karabogui.project.topo_listener import SystemTopologyListener
from karabogui.project.utils import restart_macro, run_macro
from karabogui.singletons.api import get_config, get_manager, get_topology
from karabogui.topology.api import is_macro_online
from karabogui.util import getSaveFileName, move_to_cursor

from .bases import (
    BaseProjectController, BaseProjectGroupController, ProjectControllerUiData)


class MacroInstanceController(BaseProjectController):
    """ A controller for running macro instances
    """
    # The instance ID of the running macro
    instance_id = String

    def context_menu(self, project_controller, parent=None):

        service_allowed = access_role_allowed(AccessRole.SERVICE_EDIT)
        menu = QMenu(parent)
        shutdown_action = QAction(icons.kill, 'Shutdown', menu)
        shutdown_action.triggered.connect(partial(self.shutdown,
                                                  parent=parent))
        shutdown_action.setEnabled(service_allowed)
        menu.addAction(shutdown_action)
        return menu

    def create_ui_data(self):
        icon = get_project_device_status_icon(ProxyStatus.ONLINE)
        return ProjectControllerUiData(icon=icon)

    def single_click(self, project_controller, parent=None):
        proxy = get_topology().get_device(self.instance_id)
        broadcast_event(KaraboEvent.ShowConfiguration, {'proxy': proxy})

    def _get_display_name(self):
        """Traits property getter for ``display_name``
        """
        return self.instance_id

    def info(self):
        return {'type': ProjectItemTypes.MACRO_INSTANCE,
                'deviceId': self.instance_id}

    # ----------------------------------------------------------------------
    # action handlers

    def shutdown(self, parent=None):
        get_manager().shutdownDevice(self.instance_id, showConfirm=True,
                                     parent=parent)


class MacroController(BaseProjectGroupController):
    """ A controller for MacroModel objects
    """
    # Redefine model with the correct type
    model = Instance(MacroModel)
    # An object which listens to system topology updates
    topo_listener = Instance(SystemTopologyListener)

    def context_menu(self, project_controller, parent=None):
        project_allowed = access_role_allowed(AccessRole.PROJECT_EDIT)
        service_allowed = access_role_allowed(AccessRole.SERVICE_EDIT)

        menu = QMenu(parent)
        edit_action = QAction(icons.edit, 'Edit', menu)
        edit_action.triggered.connect(partial(self._edit_macro, parent=parent))
        edit_action.setEnabled(project_allowed)

        dupe_action = QAction(icons.editCopy, 'Duplicate', menu)
        dupe_action.triggered.connect(partial(self._duplicate_macro,
                                              project_controller,
                                              parent=parent))
        dupe_action.setEnabled(project_allowed)

        delete_action = QAction(icons.kill, 'Delete', menu)
        delete_action.triggered.connect(partial(self._delete_macro,
                                                project_controller,
                                                parent=parent))
        delete_action.setEnabled(project_allowed)

        save_as_action = QAction(icons.saveAs, 'Save to file', menu)
        save_as_action.triggered.connect(partial(self._save_macro_to_file,
                                                 parent=parent))
        run_action = QAction(icons.run, 'Run', menu)
        run_action.triggered.connect(partial(self.start_macro, parent=parent))
        run_action.setEnabled(service_allowed)

        menu.addAction(edit_action)
        menu.addAction(dupe_action)
        menu.addAction(delete_action)
        menu.addSeparator()
        menu.addAction(save_as_action)
        menu.addSeparator()
        menu.addAction(run_action)
        return menu

    def create_ui_data(self):
        return ProjectControllerUiData(icon=icons.file)

    def info(self):
        return {'type': ProjectItemTypes.MACRO,
                'deviceId': self.model.instance_id,
                'code': self.model.code}

    def double_click(self, project_controller, parent=None):
        broadcast_event(KaraboEvent.ShowMacroView,
                        {'model': self.model})

    def item_handler(self, added, removed):
        """ Called when instances are added/removed
        """
        for inst_id in removed:
            controller = self._child_map[inst_id]
            # NOTE: Always use a context manager when modifying self.children!
            with self._qt_model.removal_context(controller):
                self.children.remove(controller)
            self.child_destroy(controller)
        model = self.model.clone_traits(
            traits=["instance_id", "project_name", "uuid"], copy="deep")
        additions = [self.child_create(model=model, parent=self,
                                       _qt_model=self._qt_model,
                                       instance_id=inst_id)
                     for inst_id in added]
        # Synchronize the GUI with the Traits model
        self._update_ui_children(additions)

    def system_topology_callback(self, devices, servers):
        """ This callback is called by the ``SystemTopologyListener`` object
        in the ``topo_listener`` trait.
        """
        data = {'model': self.model}
        running_instances = self.model.instances
        for dev_id, class_id, status in devices:
            if dev_id.startswith(self.model.instance_id):
                if (status is ProxyStatus.OFFLINE and
                        dev_id in running_instances):
                    running_instances.remove(dev_id)
                elif dev_id not in running_instances:
                    running_instances.append(dev_id)
                    data['instance'] = dev_id
                    # Create KaraboBroadcastEvent
                    broadcast_event(KaraboEvent.ConnectMacroInstance,
                                    data)

    # ----------------------------------------------------------------------
    # Traits handlers

    @on_trait_change("model:initialized")
    def _update_ui_instances(self):
        """ Whenever the object is modified it should be visible to the user
        """
        self.model.instances = _get_macro_instances(self.model.instance_id)
        project = self.parent
        if project is not None:
            self.model.project_name = project.model.simple_name

    # ----------------------------------------------------------------------
    # traits notification handlers

    def _children_items_changed(self, event):
        """ Maintain ``_child_map`` by watching item events on ``children``

        This is a static notification handler which is connected automatically
        by Traits.
        """
        for controller in event.removed:
            self._child_map.pop(controller.instance_id, None)

        for controller in event.added:
            self._child_map[controller.instance_id] = controller

    def _topo_listener_changed(self, name, old, new):
        """Handle broadcast event registration/unregistration here.
        """
        if old is not None:
            unregister_from_broadcasts(
                {KaraboEvent.SystemTopologyUpdate: old._event_topology})
        if new is not None:
            register_for_broadcasts(
                {KaraboEvent.SystemTopologyUpdate: new._event_topology})

    # ----------------------------------------------------------------------
    # action handlers

    def _delete_macro(self, project_controller, parent=None):
        """ Remove the macro associated with this item from its project
        """
        macro = self.model

        ask = ('Are you sure you want to delete \"<b>{}</b>\".<br /> '
               'Continue action?'.format(macro.simple_name))
        msg_box = QMessageBox(QMessageBox.Question, 'Delete macro',
                              ask, QMessageBox.Yes | QMessageBox.No,
                              parent=parent)
        msg_box.setModal(False)
        msg_box.setDefaultButton(QMessageBox.No)
        move_to_cursor(msg_box)
        if msg_box.exec() == QMessageBox.Yes:
            project = project_controller.model
            if macro in project.macros:
                project.macros.remove(macro)

            broadcast_event(KaraboEvent.RemoveProjectModelViews,
                            {'models': [macro]})

    def _edit_macro(self, parent=None):
        dialog = ObjectEditDialog(object_type='macro', model=self.model,
                                  parent=parent)
        move_to_cursor(dialog)
        result = dialog.exec()
        if result == QDialog.Accepted:
            self.model.simple_name = dialog.simple_name

    def _duplicate_macro(self, project_controller, parent=None):
        macro = self.model
        project = project_controller.model
        dialog = ObjectDuplicateDialog(macro.simple_name, parent=parent)
        move_to_cursor(dialog)
        if dialog.exec() == QDialog.Accepted:
            code = write_macro(macro)
            for simple_name in dialog.duplicate_names:
                dupe_macro = read_macro(StringIO(code))
                dupe_macro.simple_name = simple_name
                project.macros.append(dupe_macro)

    def _save_macro_to_file(self, parent=None):
        config = get_config()
        path = config['data_dir']
        directory = path if path and op.isdir(path) else ""

        macro = self.model
        filename = macro.simple_name
        filename = re.sub(r'[\W]', '-', filename)
        fn = getSaveFileName(caption='Save macro to file',
                             filter='Python Macro (*.py)',
                             suffix='py',
                             selectFile=filename,
                             directory=directory,
                             parent=parent)
        if not fn:
            return

        config['data_dir'] = op.dirname(fn)

        if not fn.endswith('.py'):
            fn = f'{fn}.py'

        with open(fn, 'w') as fout:
            fout.write(write_macro(macro))

    def start_macro(self, parent=None):
        """Run not running macro and restart running macro

        NOTE: Macro can be restarted only after user confirmation
        """
        instance_id = self.model.instance_id
        if is_macro_online(instance_id):
            restart_macro(instance_id, self.run_macro)
        else:
            self.run_macro()

    def run_macro(self, parent=None):
        """Action handler to instantiate the macro

        NOTE: This method is also used to instantiate all macros in a project!
        """
        try:
            compile(self.model.code, self.model.simple_name, "exec")
        except SyntaxError as e:
            # Show erroneous macro before showing the error dialog
            broadcast_event(KaraboEvent.ShowMacroView,
                            {'model': self.model})
            formatted_msg = "{}\n{}{}^\nin {} line {}".format(
                e.msg, e.text, " " * e.offset, e.filename, e.lineno)
            messagebox.show_warning(formatted_msg, title=type(e).__name__,
                                    parent=parent)
            return

        run_macro(self.model, parent)

# ----------------------------------------------------------------------


def _get_macro_instances(macro_id):
    instances = set()

    def visitor(node):
        attrs = node.attributes
        if (attrs.get('type') == 'macro' and
                node.node_id.startswith(macro_id)):
            instances.add(node.node_id)

    get_topology().visit_system_tree(visitor)
    return list(instances)


def get_project_macros(project_controller):
    """Given a ``project_controller`` return all the online and offline macros
    """
    online = []
    offline = []

    def visitor(obj):
        if isinstance(obj, MacroController):
            if obj.children:
                online.append(obj)
            else:
                offline.append(obj)

    walk_traits_object(project_controller, visitor)

    return online, offline



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from traits.api import Array

from ..base.analyzer import BaseAnalyzer


class HistogramAnalyzer(BaseAnalyzer):

    # The calculated edges of the input array
    _edges = Array

    # The counts of the edges from the input array
    _hist = Array

    # -----------------------------------------------------------------------
    # Public methods

    def analyze(self, region, **config):
        """Calculates the histogram of a 1D array."""
        self._hist, self._edges = np.histogram(region.flatten(), bins="sqrt")
        return self._edges, self._hist

    def clear_data(self):
        """Clear the previously calculated histogram and edges"""
        self._hist = np.array([])
        self._edges = np.array([])

    # -----------------------------------------------------------------------
    # Trait properties

    def _get_stats(self):
        """Calculates and histogram statistics"""
        stats = {}
        if self._hist.size and self._edges.size:
            # Get the bin centers to use for mean and std calculations
            centers = (self._edges[1:] + self._edges[:-1]) / 2

            stats["count"] = self._hist.sum()
            stats["mean"] = np.average(centers, weights=self._hist)
            stats["mode"] = self._edges[self._hist.argmax()]
            stats["min"] = self._edges[0]
            stats["max"] = self._edges[-1]
            stats["std"] = np.sqrt(np.average((centers - stats["mean"]) ** 2,
                                              weights=self._hist))
            stats["bins"] = len(self._edges)
            stats["bin_width"] = self._edges[1] - self._edges[0]

        return stats



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from enum import Enum

from qtpy.QtCore import QEvent
from qtpy.QtWidgets import QApplication

from karabogui.singletons.api import get_mediator


# Enum for karabo broadcast event senders
class KaraboEvent(Enum):
    AccessLevelChanged = "Global Access Level changed"
    ActiveDestinations = "Active Proposal (Logbook) list "
    BigDataProcessing = "Big data delay"
    ClearConfigurator = "Clear configurator"
    ConnectMacroInstance = "Connect macro instance"
    CreateMainWindow = "Create the main window"
    CustomEvent = "Custom Event"  # Generic event
    DatabaseIsBusy = "Database is busy"
    DeviceInitReply = "Device instantiate reply"
    ListConfigurationUpdated = "Configuration List Updated"
    LoadConfiguration = "Load device configuration"
    LoginUserChanged = "User changed"
    LogMessages = "Log messages"
    MaximizePanel = "Maximize a panel"
    MinimizePanel = "Minimize a panel"
    MiddlePanelClosed = "Middle panel closed"
    NetworkConnectStatus = "Network connection changed"
    OpenSceneLink = "Scene link"
    RemoveProjectModelViews = "Remove scene or macro views"
    ProjectName = "Update main window project name"
    ProjectDomainsList = "Project domains list"
    ProjectTrashed = "Project Trashed"
    ProjectFilterUpdated = "Project filter updated"
    ProjectItemsList = "Project items list"
    ProjectItemsLoaded = "Project items loaded"
    ProjectItemsSaved = "Project items saved"
    ProjectUpdated = "Project Updated"
    ProjectFindWithItem = "Find projects with device"
    RaiseEditor = "Raise the Editor"
    RestoreSceneView = "Restore the scene"
    ServerInformationUpdate = "Server information update"
    ServerNotification = "Server Notification"
    ShowConfiguration = "Show configuration"
    ShowInitConfiguration = "Show Configuration From Name"
    ShowConfigurationFromPast = "Show Configuration From Past"
    ShowDevice = "Show device"
    ShowProjectDevice = "Show project device"
    ShowProjectModel = "Show project model"
    ShowMacroView = "Show macro"
    ShowSceneView = "Show scene"
    ShowUnattachedController = "Show unattached controller"
    ShowUnattachedSceneView = "Show unattached scene"
    StartMonitoringDevice = "Start listening to device updates"
    StopMonitoringDevice = "Stop listening to device updates"
    SystemTopologyUpdate = "System topology updated"
    SystemTopologyInstanceUpdate = "System topology instance(device) updated"
    UserSession = "User session"
    UserSessionInfo = "User session info"
    UpdateDeviceConfigurator = "Update device in configurator"
    UpdateValueConfigurator = "Value update in configurator"


class KaraboBroadcastEvent(QEvent):
    """ Custom event to handle GUI widget communication """

    # Ask Qt for our event type
    Type = QEvent.Type(QEvent.registerEventType())

    def __init__(self, sender="", data=None):
        super().__init__(self.Type)
        self.sender = sender  # Names the sender
        self.data = data or {}  # Includes the data which is sent


def broadcast_event(sender_enum, data):
    """ Broadcast the given `event`.
    """
    assert isinstance(sender_enum, KaraboEvent)

    mediator = get_mediator()
    event = KaraboBroadcastEvent(sender=sender_enum, data=data)
    QApplication.postEvent(mediator, event)


def register_for_broadcasts(event_map):
    """ Register the given `qobject` to the events coming from the singleton
        `mediator`.
    """
    get_mediator().register_listener(event_map)


def unregister_from_broadcasts(event_map):
    """ Unregister the given `qobject` from the events coming from the
        singleton mediator object.
    """
    get_mediator().unregister_listener(event_map)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on June 2, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QSortFilterProxyModel, Qt


class ConfiguratorFilterModel(QSortFilterProxyModel):

    def __init__(self, source_model=None, parent=None):
        super().__init__(parent)
        self.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.setFilterRole(Qt.DisplayRole)
        self.setFilterKeyColumn(0)
        self.setRecursiveFilteringEnabled(True)
        self.setDynamicSortFilter(False)
        self.setSourceModel(source_model)

    @property
    def root(self):
        """Return the root proxy of the source model"""
        return self.sourceModel().root

    # --------------------------------------------------------------------
    # Index Methods

    def index_ref(self, index):
        source_index = self.mapToSource(index)
        if not source_index.isValid():
            return None
        return self.sourceModel().index_ref(source_index)

    def indexInfo(self, index):
        source_index = self.mapToSource(index)
        if not source_index.isValid():
            return {}
        return self.sourceModel().indexInfo(source_index)

    def clear_index_modification(self, index):
        index = self.mapToSource(index)
        return self.sourceModel().clear_index_modification(index)

    def flush_index_modification(self, index):
        index = self.mapToSource(index)
        return self.sourceModel().flush_index_modification(index)
